"""Test array tools."""

from __future__ import annotations

import warnings

import numpy as np
import pytest
import rasterio as rio

import geoutils as gu
from geoutils.raster.array import get_array_and_mask, get_valid_extent, get_xy_rotated


class TestArray:
    @pytest.mark.parametrize("dtype", ["uint8", "uint16", "int32", "float32", "float16"])  # type: ignore
    @pytest.mark.parametrize(
        "mask_and_viewable",
        [
            (None, True),  # A ndarray with no mask should support views
            (False, True),  # A masked array with an empty mask should support views
            ([True, False, False, False], False),  # A masked array with an occupied mask should not support views.
            ([False, False, False, False], True),  # A masked array with an empty occupied mask should support views.
        ],
    )  # type: ignore
    @pytest.mark.parametrize(
        "shape_and_check_passes",
        [
            ((1, 2, 2), True),  # A 3D array with a shape[0] == 1 is okay.
            ((2, 1, 2), False),  # A 3D array with a shape[0] != 1 is not okay.
            ((2, 2), True),  # A 2D array is okay.
            ((4,), True),  # A 1D array is okay.
        ],
    )  # type: ignore
    def test_get_array_and_mask(
        self,
        dtype: str,
        mask_and_viewable: tuple[None | bool | list[bool], bool],
        shape_and_check_passes: tuple[tuple[int, ...], bool],
    ) -> None:
        """Validate that the function returns views when expected, and copies otherwise."""
        warnings.simplefilter("error")

        masked_values, view_should_be_possible = mask_and_viewable
        shape, check_should_pass = shape_and_check_passes

        # Create an array of the specified dtype
        array = np.ones(shape, dtype=dtype)
        if masked_values is not None:
            if masked_values is False:
                array = np.ma.masked_array(array)
            else:
                array = np.ma.masked_array(array, mask=np.reshape(masked_values, array.shape))

        # Validate that incorrect shapes raise the correct error.
        if not check_should_pass:
            with pytest.raises(ValueError, match="Invalid array shape given"):
                get_array_and_mask(array, check_shape=True)

            # Stop the test here as the failure is now validated.
            return

        # Get a copy of the array and check its shape (it should always pass at this point)
        arr, _ = get_array_and_mask(array, copy=True, check_shape=True)

        # Validate that the array is a copy
        assert not np.shares_memory(arr, array)

        # If it was an integer dtype and it had a mask, validate that the array is now "float32"
        if np.issubdtype(dtype, np.integer) and np.any(masked_values or False):
            assert arr.dtype == "float32"

        # If there was no mask or the mask was empty, validate that arr and array are equivalent
        if not np.any(masked_values or False):
            assert np.sum(np.abs(array - arr)) == 0.0

        with warnings.catch_warnings(record=True) as caught_warnings:
            warnings.simplefilter("always")

            # Try to create a view.
            arr_view, mask = get_array_and_mask(array, copy=False)

            # If it should be possible, validate that there were no warnings.
            if view_should_be_possible:
                assert len(caught_warnings) == 0, (caught_warnings[0].message, array)
            # Otherwise, validate that one warning was raised with the correct text.
            else:
                assert len(caught_warnings) == 1
                assert "Copying is required" in str(caught_warnings[0].message)

        # Validate that the view shares memory if it was possible, or otherwise that it is a copy.
        if view_should_be_possible:
            assert np.shares_memory(array, arr_view)
        else:
            assert not np.shares_memory(array, arr_view)

    def test_get_valid_extent(self) -> None:
        """Check the function to get valid extent."""

        # Create an artificial array and masked array
        arr = np.ones(shape=(5, 5))
        arr_mask = np.zeros(shape=(5, 5), dtype=bool)
        mask_ma = np.ma.masked_array(data=arr, mask=arr_mask)

        # For no invalid values, the function should return the edges
        # For the array
        assert (0, 4, 0, 4) == get_valid_extent(arr)
        # For the masked-array
        assert (0, 4, 0, 4) == get_valid_extent(mask_ma)

        # 1/ First column:
        # If we mask it in the masked array
        mask_ma[0, :] = np.ma.masked
        assert (1, 4, 0, 4) == get_valid_extent(mask_ma)

        # If we changed the array to NaNs
        arr[0, :] = np.nan
        assert (1, 4, 0, 4) == get_valid_extent(arr)
        mask_ma.data[0, :] = np.nan
        mask_ma.mask = False
        assert (1, 4, 0, 4) == get_valid_extent(mask_ma)

        # 2/ First row:
        arr = np.ones(shape=(5, 5))
        arr_mask = np.zeros(shape=(5, 5), dtype=bool)
        mask_ma = np.ma.masked_array(data=arr, mask=arr_mask)
        # If we mask it in the masked array
        mask_ma[:, 0] = np.ma.masked
        assert (0, 4, 1, 4) == get_valid_extent(mask_ma)

        # If we changed the array to NaNs
        arr[:, 0] = np.nan
        assert (0, 4, 1, 4) == get_valid_extent(arr)
        mask_ma.data[:, 0] = np.nan
        mask_ma.mask = False
        assert (0, 4, 1, 4) == get_valid_extent(mask_ma)

        # 3/ Last column:
        arr = np.ones(shape=(5, 5))
        arr_mask = np.zeros(shape=(5, 5), dtype=bool)
        mask_ma = np.ma.masked_array(data=arr, mask=arr_mask)

        # If we mask it in the masked array
        mask_ma[-1, :] = np.ma.masked
        assert (0, 3, 0, 4) == get_valid_extent(mask_ma)

        # If we changed the array to NaNs
        arr[-1, :] = np.nan
        assert (0, 3, 0, 4) == get_valid_extent(arr)
        mask_ma.data[-1, :] = np.nan
        mask_ma.mask = False
        assert (0, 3, 0, 4) == get_valid_extent(mask_ma)

        # 4/ Last row:
        arr = np.ones(shape=(5, 5))
        arr_mask = np.zeros(shape=(5, 5), dtype=bool)
        mask_ma = np.ma.masked_array(data=arr, mask=arr_mask)

        # If we mask it in the masked array
        mask_ma[:, -1] = np.ma.masked
        assert (0, 4, 0, 3) == get_valid_extent(mask_ma)

        # If we changed the array to NaNs
        arr[:, -1] = np.nan
        assert (0, 4, 0, 3) == get_valid_extent(arr)
        mask_ma.data[:, -1] = np.nan
        mask_ma.mask = False
        assert (0, 4, 0, 3) == get_valid_extent(mask_ma)

    def test_get_xy_rotated(self) -> None:
        """Check the function to rotate array."""

        # Create an artificial raster
        rng = np.random.default_rng(42)
        width = height = 5
        transform = rio.transform.from_bounds(0, 0, 1, 1, width, height)
        r1 = gu.Raster.from_array(rng.integers(1, 255, (height, width), dtype="uint8"), transform=transform, crs=None)

        # First, we get initial coords
        xx, yy = r1.coords(grid=True, force_offset="ll")

        # Rotating the coordinates 90 degrees should be the same as rotating the array
        xx90, yy90 = get_xy_rotated(r1, along_track_angle=90)
        assert np.allclose(np.rot90(xx90), xx)
        assert np.allclose(np.rot90(yy90), yy)

        # Same for 180 degrees
        xx180, yy180 = get_xy_rotated(r1, along_track_angle=180)
        assert np.allclose(np.rot90(xx180, k=2), xx)
        assert np.allclose(np.rot90(yy180, k=2), yy)

        # Same for 270 degrees
        xx270, yy270 = get_xy_rotated(r1, along_track_angle=270)
        assert np.allclose(np.rot90(xx270, k=3), xx)
        assert np.allclose(np.rot90(yy270, k=3), yy)

        # 360 degrees should get us back on our feet
        xx360, yy360 = get_xy_rotated(r1, along_track_angle=360)
        assert np.allclose(xx360, xx)
        assert np.allclose(yy360, yy)

        # Test that the values make sense for 45 degrees
        xx45, yy45 = get_xy_rotated(r1, along_track_angle=45)
        # Should have zero on the upper left corner for xx
        assert xx45[0, 0] == pytest.approx(0)
        # Then a multiple of sqrt2 along each dimension
        assert xx45[1, 0] == pytest.approx(xx45[0, 1]) == pytest.approx(0.1 * np.sqrt(2))
        # The lower right corner should have the highest coordinate (0.8) times sqrt(2)
        assert xx45[-1, -1] == pytest.approx(np.max(xx) * np.sqrt(2))

        # Finally, yy should be rotated by 90
        assert np.allclose(np.rot90(xx45), yy45)

        xx, yy = get_xy_rotated(r1, along_track_angle=90)



"""Tests for dask-delayed functions."""

from __future__ import annotations

import os
import sys
from tempfile import NamedTemporaryFile
from typing import Any, Callable

import dask.array as da
import numpy as np
import pandas as pd
import pytest
import rasterio as rio
import xarray as xr
from dask.distributed import Client, LocalCluster
from dask_memusage import install
from pluggy import PluggyTeardownRaisedWarning
from pyproj import CRS

from geoutils.examples import _EXAMPLES_DIRECTORY
from geoutils.raster.delayed import (
    delayed_interp_points,
    delayed_reproject,
    delayed_subsample,
)

# Ignore teardown warning given by Dask when closing the local cluster (due to dask-memusage plugin)
pytestmark = pytest.mark.filterwarnings("ignore", category=PluggyTeardownRaisedWarning)


@pytest.fixture(scope="module")  # type: ignore
def cluster():
    """Fixture to use a single cluster for the entire module (otherwise raise runtime errors)."""
    # Need cluster to be single-threaded to use dask-memusage confidently
    dask_cluster = LocalCluster(n_workers=1, threads_per_worker=1, dashboard_address=None)
    yield dask_cluster
    dask_cluster.close()


def _run_dask_measuring_memusage(
    cluster: Any, dask_func: Callable[..., Any], *args_dask_func: Any, **kwargs_dask_func: Any
) -> tuple[Any, float]:
    """Run a dask function monitoring its memory usage."""

    # Create a name temporary file that won't delete immediately
    fn_tmp_csv = NamedTemporaryFile(suffix=".csv", delete=False).name

    # Setup cluster and client within context managers for a clean shutdown
    install(cluster.scheduler, fn_tmp_csv)
    with Client(cluster) as _:
        outputs = dask_func(*args_dask_func, **kwargs_dask_func)

    # Read memusage file and cleanup
    df = pd.read_csv(fn_tmp_csv)
    os.remove(fn_tmp_csv)

    # Keep only non-zero memory usage
    ind_nonzero = df.max_memory_mb != 0

    # Compute peak additional memory usage from min baseline
    memusage_mb = np.max(df.max_memory_mb[ind_nonzero]) - np.min(df.max_memory_mb[ind_nonzero])

    return outputs, memusage_mb


def _estimate_subsample_memusage(darr: da.Array, chunksizes_in_mem: tuple[int, int], subsample_size: int) -> float:
    """
    Estimate the theoretical memory usage of the delayed subsampling method.
    (we don't need to be super precise, just within a factor of ~2 to check memory usage performs as expected)
    """

    # TOTAL SIZE = Single chunk operations + Subsample indexes + Metadata passed to dask + Outputs

    # On top of the rest is added the Dask graph, we will multiply by a factor of 2.5 to get a good safety margin
    fac_dask_margin = 2.5
    num_chunks = np.prod(darr.numblocks)

    # Single chunk operation = (data type bytes + boolean from np.isfinite) * chunksize **2
    chunk_memusage = (darr.dtype.itemsize + np.dtype("bool").itemsize) * np.prod(chunksizes_in_mem)

    # 1D index subsample size: integer type * subsample_size
    sample_memusage = np.dtype("int32").itemsize * subsample_size

    # Outputs: number of valid pixels + subsample
    valids_memusage = np.dtype("int32").itemsize * num_chunks
    subout_memusage = np.dtype(darr.dtype).itemsize * subsample_size
    out_memusage = valids_memusage + subout_memusage

    # Size of metadata passed to dask: number of blocks times its content
    # Content of a metadata block = list (block size) of list (subsample size) of integer indexes
    size_index_int = 28  # Python size for int
    list_all_blocks = 64 + 8 * num_chunks  # A list is 64 + 8 bits per element, without memory of contained elements
    list_per_block = 64 * num_chunks + 8 * subsample_size + size_index_int * subsample_size  # Rough max estimate
    meta_memusage = list_per_block + list_all_blocks

    # Final estimate of memory usage of operation in MB
    max_op_memusage = fac_dask_margin * (chunk_memusage + sample_memusage + out_memusage + meta_memusage) / (2**20)
    # We add a base memory usage of ~80 MB + 10MB per 1000 chunks (loaded in background by Dask even on tiny data)
    max_op_memusage += 80 + 10 * (num_chunks / 1000)

    return max_op_memusage


def _estimate_interp_points_memusage(darr: da.Array, chunksizes_in_mem: tuple[int, int], ninterp: int) -> float:
    """
    Estimate the theoretical memory usage of the delayed interpolation method.
    (we don't need to be super precise, just within a factor of ~2 to check memory usage performs as expected)
    """

    # TOTAL SIZE = Single chunk operations + Chunk overlap + Metadata passed to dask + Outputs

    # On top of the rest is added the Dask graph, we will multiply by a factor of 2.5 to get a good safety margin
    fac_dask_margin = 2.5
    num_chunks = np.prod(darr.numblocks)

    # Single chunk operation = (data type bytes + boolean from np.isfinite + its subset) * overlapping chunksize **2
    chunk_memusage = (darr.dtype.itemsize + 2 * np.dtype("bool").itemsize) * np.prod(chunksizes_in_mem)
    # For interpolation, chunks have to overlap and temporarily load each neighbouring chunk,
    # we add 8 neighbouring chunks, and double the size due to the memory used during interpolation
    chunk_memusage *= 9

    # Outputs: pair of interpolated coordinates
    out_memusage = np.dtype(darr.dtype).itemsize * ninterp * 2

    # Size of metadata passed to dask: number of blocks times its content
    # Content of a metadata block = list (block size) of list (subsample size) of integer
    size_index_int = 28  # Python size for int
    size_index_float = 24  # Python size for float
    list_all_blocks = 64 + 8 * num_chunks  # A list is 64 + 8 bits per element, without memory of contained elements
    list_per_block = 64 * num_chunks + 8 * ninterp + size_index_int * ninterp  # Rough max estimate
    # And a list for each block of dict with 4 floats (xres, yres, xstart, ystart)
    dict_all_blocks = 64 * num_chunks + 4 * size_index_float * 64 * num_chunks
    meta_memusage = list_per_block + list_all_blocks + dict_all_blocks

    # Final estimate of memory usage of operation in MB
    max_op_memusage = fac_dask_margin * (chunk_memusage + out_memusage + meta_memusage) / (2**20)
    # We add a base memory usage of ~80 MB + 10MB per 1000 chunks (loaded in background by Dask even on tiny data)
    max_op_memusage += 80 + 10 * (num_chunks / 1000)

    return max_op_memusage


def _estimate_reproject_memusage(
    darr: da.Array,
    chunksizes_in_mem: tuple[int, int],
    dst_chunksizes: tuple[int, int],
    rel_res_fac: tuple[float, float],
) -> float:
    """
    Estimate the theoretical memory usage of the delayed reprojection method.
    (we don't need to be super precise, just within a factor of ~2 to check memory usage performs as expected)
    """

    # TOTAL SIZE = Combined source chunk operations + Building geopandas mapping + Metadata passed to dask + Outputs

    # On top of the rest is added the Dask graph, we will multiply by a factor of 2.5 to get a good safety margin
    fac_dask_margin = 2.5
    num_chunks = np.prod(darr.numblocks)

    # THE BIG QUESTION: how many maximum source chunks might be loaded for a single destination chunk?
    # It depends on the relative ratio of input chunksizes to destination chunksizes, accounting for resolution change
    x_rel_source_chunks = dst_chunksizes[0] / chunksizes_in_mem[0] * rel_res_fac[0]
    y_rel_source_chunks = dst_chunksizes[1] / chunksizes_in_mem[1] * rel_res_fac[1]
    # There is also some overlap needed for resampling and due to warping in different CRS, so let's multiply this by 8
    # (all neighbouring tiles)
    nb_source_chunks_per_dest = 8 * x_rel_source_chunks * y_rel_source_chunks

    # Combined memory usage of one chunk operation = squared array made from combined chunksize + original chunks
    total_nb = np.ceil(np.sqrt(nb_source_chunks_per_dest)) ** 2 + nb_source_chunks_per_dest
    # We multiply the memory usage of a single chunk to the number of loaded/combined chunks
    chunk_memusage = darr.dtype.itemsize * np.prod(chunksizes_in_mem) * total_nb

    # Outputs: reprojected raster
    out_memusage = np.dtype(darr.dtype).itemsize * np.prod(dst_chunksizes)

    # Size of metadata passed to dask: number of blocks times its content
    # For each block, we pass a dict with 4 integers per source chunk (rxs, rxe, rys, rye)
    size_index_float = 24  # Python size for float
    size_index_int = 28  # Python size for float
    dict_all_blocks = (64 + 4 * size_index_int * nb_source_chunks_per_dest) * num_chunks
    # Passing the 2 CRS, 2 transforms, resampling methods and 2 nodatas
    combined_meta = (112 + 112 + 56 + 56 + 44 + 28 + 28) * size_index_float * num_chunks
    meta_memusage = combined_meta + dict_all_blocks

    # Final estimate of memory usage of operation in MB
    max_op_memusage = fac_dask_margin * (chunk_memusage + out_memusage + meta_memusage) / (2**20)
    # We add a base memory usage of ~80 MB + 10MB per 1000 chunks (loaded in background by Dask even on tiny data)
    max_op_memusage += 80 + 10 * (num_chunks / 1000)

    return max_op_memusage


def _build_dst_transform_shifted_newres(
    src_transform: rio.transform.Affine,
    src_shape: tuple[int, int],
    src_crs: CRS,
    dst_crs: CRS,
    bounds_rel_shift: tuple[float, float],
    res_rel_fac: tuple[float, float],
) -> rio.transform.Affine:
    """
    Build a destination transform intersecting the source transform given source/destination shapes,
    and possibly introducing a relative shift in upper-left bound and multiplicative change in resolution.
    """

    # Get bounding box in source CRS
    bounds = rio.coords.BoundingBox(*rio.transform.array_bounds(src_shape[0], src_shape[1], src_transform))

    # Construct an aligned transform in the destination CRS assuming the same resolution
    tmp_transform = rio.warp.calculate_default_transform(
        src_crs,
        dst_crs,
        src_shape[1],
        src_shape[0],
        left=bounds.left,
        right=bounds.right,
        top=bounds.top,
        bottom=bounds.bottom,
        dst_width=src_shape[1],
        dst_height=src_shape[0],
    )[0]
    # This allows us to get bounds and resolution in the units of the new CRS
    tmp_res = (tmp_transform[0], abs(tmp_transform[4]))
    tmp_bounds = rio.coords.BoundingBox(*rio.transform.array_bounds(src_shape[0], src_shape[1], tmp_transform))
    # Now we can create a shifted/different-res destination grid
    dst_transform = rio.transform.from_origin(
        west=tmp_bounds.left + bounds_rel_shift[0] * tmp_res[0] * src_shape[1],
        north=tmp_bounds.top + 150 * bounds_rel_shift[0] * tmp_res[1] * src_shape[0],
        xsize=tmp_res[0] / res_rel_fac[0],
        ysize=tmp_res[1] / res_rel_fac[1],
    )

    return dst_transform


class TestDelayed:
    """
    Testing class for delayed functions.

    We test on a first set of rasters big enough to clearly monitor the memory usage, and a second set small enough
    to run fast to check a wide range of input parameters.

    In details:
    Set 1. We capture memory usage during the .compute() calls and check that only the expected amount of memory that
        we estimate independently (bytes used by one or several chunk combinations + metadata) is indeed used.
    Set 2. We compare outputs with the in-memory function specifically for input variables that influence the delayed
        algorithm and might lead to new errors (for example: array shape to get subsample/points locations for
        subsample and interp_points, or destination chunksizes to map output of reproject).

    We start with set 2: output checks which run faster when ordered before
    (maybe due to the cluster memory monitoring after).
    """

    # Define random seed for generating test data
    rng = da.random.default_rng(seed=42)

    # 1/ Set 1: Memory usage checks

    # Big test files written on disk in an out-of-memory fashion,
    # with different input shapes not necessarily aligned between themselves
    large_shape = (10000, 10000)
    # We can use a constant value for storage chunks, as it doesn't have any influence on the accuracy of delayed
    # methods (can change slightly RAM usage, but pretty stable as long as chunksizes in memory are larger and
    # significantly bigger)
    chunksizes_on_disk = (500, 500)
    fn_large = os.path.join(_EXAMPLES_DIRECTORY, "test_large.nc")
    if not os.path.exists(fn_large):
        # Create random array in the right shape
        data = rng.normal(size=large_shape[0] * large_shape[1]).reshape(large_shape[0], large_shape[1])
        data_arr = xr.DataArray(data=data, dims=["x", "y"])
        ds = xr.Dataset(data_vars={"test": data_arr})
        encoding_kwargs = {"test": {"chunksizes": chunksizes_on_disk}}
        # Write to disk out-of-memory
        writer = ds.to_netcdf(fn_large, encoding=encoding_kwargs, compute=False)
        writer.compute()

    # 2. Set 2
    # Smaller test files for fast checks, with various shapes and with/without nodata
    list_small_shapes = [(50, 50), (51, 47)]
    with_nodata = [False, True]
    list_small_darr = []
    for small_shape in list_small_shapes:
        for w in with_nodata:
            small_darr = rng.normal(size=small_shape[0] * small_shape[1])
            # Add about half nodata values
            if w:
                ind_nodata = rng.choice(small_darr.size, size=int(small_darr.size / 2), replace=False)
                small_darr[list(ind_nodata)] = np.nan
            small_darr = small_darr.reshape(small_shape[0], small_shape[1])
            list_small_darr.append(small_darr)

    # List of in-memory chunksize for small tests
    list_small_chunksizes_in_mem = [(10, 10), (7, 19)]

    # Create a corresponding boolean array for each numerical dask array
    # Every finite numerical value (valid numerical value) corresponds to True (valid boolean value).
    darr_bool = []
    for small_darr in list_small_darr:
        darr_bool.append(da.where(da.isfinite(small_darr), True, False))

    @pytest.mark.parametrize("darr, darr_bool", list(zip(list_small_darr, darr_bool)))  # type: ignore
    @pytest.mark.parametrize("chunksizes_in_mem", list_small_chunksizes_in_mem)  # type: ignore
    @pytest.mark.parametrize("subsample_size", [2, 100, 100000])  # type: ignore
    def test_delayed_subsample__output(
        self, darr: da.Array, darr_bool: da.Array, chunksizes_in_mem: tuple[int, int], subsample_size: int
    ):
        """
        Checks for delayed subsampling function for output accuracy.
        Variables that influence specifically the delayed function and might lead to new errors are:
        - Input chunksizes,
        - Input array shape,
        - Number of subsampled points.
        """

        # 1/ We run the delayed function after re-chunking
        darr = darr.rechunk(chunksizes_in_mem)
        sub = delayed_subsample(darr, subsample=subsample_size, random_state=42)
        # 2/ Output checks

        # # The subsample should have exactly the prescribed length, with only valid values
        assert len(sub) == min(subsample_size, np.count_nonzero(np.isfinite(darr)))
        assert all(np.isfinite(sub))

        # To verify the sampling works correctly, we can get its subsample indices with the argument return_indices
        # And compare to the same subsample with vindex (now that we know the coordinates of valid values sampled)
        indices = delayed_subsample(darr, subsample=subsample_size, random_state=42, return_indices=True)
        sub2 = np.array(darr.vindex[indices[0], indices[1]])
        assert np.array_equal(sub, sub2)

        # Finally, to verify that a boolean array, with valid values at the same locations as the numerical array,
        # leads to the same results, we compare the samples values and the samples indices.
        darr_bool = darr_bool.rechunk(chunksizes_in_mem)
        indices_bool = delayed_subsample(darr_bool, subsample=subsample_size, random_state=42, return_indices=True)
        sub_bool = np.array(darr.vindex[indices_bool])
        assert np.array_equal(sub, sub_bool)
        assert np.array_equal(indices, indices_bool)

    @pytest.mark.parametrize("darr", list_small_darr)  # type: ignore
    @pytest.mark.parametrize("chunksizes_in_mem", list_small_chunksizes_in_mem)  # type: ignore
    @pytest.mark.parametrize("ninterp", [2, 100])  # type: ignore
    @pytest.mark.parametrize("res", [(0.5, 2), (1, 1)])  # type: ignore
    def test_delayed_interp_points__output(
        self, darr: da.Array, chunksizes_in_mem: tuple[int, int], ninterp: int, res: tuple[float, float]
    ):
        """
        Checks for delayed interpolate points function.
        Variables that influence specifically the delayed function are:
        - Input chunksizes,
        - Input array shape,
        - Number of interpolated points,
        - The resolution of the regular grid.
        """

        # 1/ Define points to interpolate given the size and resolution
        darr = darr.rechunk(chunksizes_in_mem)
        rng = np.random.default_rng(seed=42)
        interp_x = (rng.choice(darr.shape[0], ninterp) + rng.random(ninterp)) * res[0]
        interp_y = (rng.choice(darr.shape[1], ninterp) + rng.random(ninterp)) * res[1]

        interp1 = delayed_interp_points(darr, points=(interp_x, interp_y), resolution=res)

        # 2/ Output checks

        # Interpolate directly with Xarray (loads a lot in memory) and check results are exactly the same
        xx = xr.DataArray(interp_x, dims="z", name="x")
        yy = xr.DataArray(interp_y, dims="z", name="y")
        ds = xr.DataArray(
            data=darr,
            dims=["x", "y"],
            coords={
                "x": np.arange(0, darr.shape[0] * res[0], res[0]),
                "y": np.arange(0, darr.shape[1] * res[1], res[1]),
            },
        )
        interp2 = ds.interp(x=xx, y=yy)
        interp2.compute()
        interp2 = np.array(interp2.values)

        assert np.array_equal(interp1, interp2, equal_nan=True)

    @pytest.mark.parametrize("darr", list_small_darr)  # type: ignore
    @pytest.mark.parametrize("chunksizes_in_mem", list_small_chunksizes_in_mem)  # type: ignore
    @pytest.mark.parametrize("dst_chunksizes", list_small_chunksizes_in_mem)  # type: ignore
    # Shift upper left corner of output bounds (relative to projected input bounds) by fractions of the raster size
    @pytest.mark.parametrize("dst_bounds_rel_shift", [(0, 0), (-0.2, 0.5)])  # type: ignore
    # Modify output resolution (relative to projected input resolution) by a factor
    @pytest.mark.parametrize("dst_res_rel_fac", [(1, 1), (2.1, 0.54)])  # type: ignore
    # Same for shape
    @pytest.mark.parametrize("dst_shape_diff", [(0, 0), (-28, 117)])  # type: ignore
    def test_delayed_reproject__output(
        self,
        darr: da.Array,
        chunksizes_in_mem: tuple[int, int],
        dst_chunksizes: tuple[int, int],
        dst_bounds_rel_shift: tuple[float, float],
        dst_res_rel_fac: tuple[float, float],
        dst_shape_diff: tuple[int, int],
    ):
        """
        Checks for the delayed reproject function.
        Variables that influence specifically the delayed function are:
        - Input/output chunksizes,
        - Input array shape,
        - Output geotransform relative to projected input geotransform,
        - Output array shape relative to input.
        """

        # Keeping this commented here if we need to redo local tests due to Rasterio errors
        # darr = list_small_darr[0]
        # chunksizes_in_mem = list_small_chunksizes_in_mem[0]
        # dst_chunksizes = list_small_chunksizes_in_mem[0]  # (2000, 2000)
        # dst_bounds_rel_shift = (0, 0)
        # dst_res_rel_fac = (0.45, 0.45)  # (1, 1)
        # dst_shape_diff = (0, 0)
        # cluster = LocalCluster(n_workers=1, threads_per_worker=1, dashboard_address=None)

        # 0/ Define input parameters

        # Get input and output shape
        darr = darr.rechunk(chunksizes_in_mem)
        src_shape = darr.shape
        dst_shape = (src_shape[0] + dst_shape_diff[0], src_shape[1] + dst_shape_diff[1])

        # Define arbitrary input transform, as we only care about the relative difference with the output transform
        src_transform = rio.transform.from_bounds(10, 10, 15, 15, src_shape[0], src_shape[1])

        # Other arguments having (normally) no influence
        src_crs = CRS(4326)
        dst_crs = CRS(32630)
        src_nodata = -9999
        dst_nodata = 99999
        resampling = rio.enums.Resampling.bilinear

        # Get shifted dst_transform with new resolution
        dst_transform = _build_dst_transform_shifted_newres(
            src_transform=src_transform,
            src_crs=src_crs,
            dst_crs=dst_crs,
            src_shape=src_shape,
            bounds_rel_shift=dst_bounds_rel_shift,
            res_rel_fac=dst_res_rel_fac,
        )

        # 2/ Run delayed reproject with memory monitoring

        reproj_arr = delayed_reproject(
            darr,
            src_transform=src_transform,
            src_crs=src_crs,
            dst_transform=dst_transform,
            dst_crs=dst_crs,
            dst_shape=dst_shape,
            src_nodata=src_nodata,
            dst_nodata=dst_nodata,
            resampling=resampling,
            dst_chunksizes=dst_chunksizes,
        )

        # 3/ Outputs check: load in memory and compare with a direct Rasterio reproject
        reproj_arr = np.array(reproj_arr)

        dst_arr = np.zeros(dst_shape)
        _ = rio.warp.reproject(
            np.array(darr),
            dst_arr,
            src_transform=src_transform,
            src_crs=src_crs,
            dst_transform=dst_transform,
            dst_crs=dst_crs,
            resampling=resampling,
            src_nodata=src_nodata,
            dst_nodata=dst_nodata,
        )

        # Keeping this to visualize Rasterio resampling issue
        # if PLOT:
        # import matplotlib.pyplot as plt
        # plt.figure()
        # plt.imshow((reproj_arr - dst_arr), cmap="RdYlBu", vmin=-0.2, vmax=0.2, interpolation="None")
        # plt.colorbar()
        # plt.savefig("/home/atom/ongoing/diff_close_zero.png", dpi=500)
        # plt.figure()
        # plt.imshow(np.abs(reproj_arr - dst_arr), cmap="RdYlBu", vmin=99997, vmax=100001, interpolation="None")
        # plt.colorbar()
        # plt.savefig("/home/atom/ongoing/diff_nodata.png", dpi=500)
        # plt.figure()
        # plt.imshow(dst_arr, cmap="RdYlBu", vmin=-1, vmax=1, interpolation="None")
        # plt.colorbar()
        # plt.savefig("/home/atom/ongoing/dst.png", dpi=500)

        # Due to (what appears to be) Rasterio errors, we have to remain imprecise for the checks here:
        # even though some reprojections are pretty good, some can get a bit nasty

        # Check that little data (less than 10% of pixels) are significantly different
        ind_signif_diff = np.abs(reproj_arr - dst_arr) > 0.5
        assert np.count_nonzero(ind_signif_diff) < 0.1 * reproj_arr.size

        # The median difference should be negligible compared to the amplitude of the signal (+/- 1 std)
        assert np.nanmedian(np.abs(reproj_arr - dst_arr)) < 0.1

        # # Replace with allclose once Rasterio issue fixed? For some cases we get a good match
        # (less than 0.01 for all pixels)
        # assert np.allclose(reproj_arr[~ind_both_nodata], dst_arr[~ind_both_nodata], atol=0.02)

    @pytest.mark.parametrize("fn", [fn_large])  # type: ignore
    @pytest.mark.parametrize("chunksizes_in_mem", [(1000, 1000), (2500, 2500)])  # type: ignore
    @pytest.mark.parametrize("subsample_size", [100, 100000])  # type: ignore
    def test_delayed_subsample__memusage(
        self, fn: str, chunksizes_in_mem: tuple[int, int], subsample_size: int, cluster: Any
    ):
        """
        Checks for delayed subsampling function for memory usage on big file.
        (and also runs output checks as not long or too memory intensive in this case)
        Variables that influence memory usage are:
        - Subsample sizes,
        - Chunksizes in memory.
        """

        # Only check on linux
        if sys.platform == "linux":

            # 0/ Open dataset with chunks
            ds = xr.open_dataset(fn, chunks={"x": chunksizes_in_mem[0], "y": chunksizes_in_mem[1]})
            darr = ds["test"].data

            # 1/ Estimation of theoretical memory usage of the subsampling script

            max_op_memusage = _estimate_subsample_memusage(
                darr=darr, chunksizes_in_mem=chunksizes_in_mem, subsample_size=subsample_size
            )

            # 2/ Run delayed subsample with dask memory usage monitoring

            # Derive subsample from delayed function
            # (passed to wrapper function to measure memory usage during execution)
            sub, measured_op_memusage = _run_dask_measuring_memusage(
                cluster, delayed_subsample, darr, subsample=subsample_size, random_state=42
            )

            # Check the measured memory usage is smaller than the maximum estimated one
            assert measured_op_memusage < max_op_memusage

            # 3/ Output checks
            # The subsample should have exactly the prescribed length, with only valid values
            assert len(sub) == subsample_size
            assert all(np.isfinite(sub))

            # To verify the sampling works correctly, we can get its subsample indices with the argument return_indices
            # And compare to the same subsample with vindex (now that we know the coordinates of valid values sampled)
            indices = delayed_subsample(darr, subsample=subsample_size, random_state=42, return_indices=True)
            sub2 = np.array(darr.vindex[indices[0], indices[1]])
            assert np.array_equal(sub, sub2)

    @pytest.mark.parametrize("fn", [fn_large])  # type: ignore
    @pytest.mark.parametrize("chunksizes_in_mem", [(2000, 2000)])  # type: ignore
    @pytest.mark.parametrize("ninterp", [100, 100000])  # type: ignore
    def test_delayed_interp_points__memusage(
        self, fn: str, chunksizes_in_mem: tuple[int, int], ninterp: int, cluster: Any
    ):
        """
        Checks for delayed interpolate points function for memory usage on a big file.
        Variables that influence memory usage are:
        - Number of interpolated points,
        - Chunksizes in memory.
        """

        # Only check on linux
        if sys.platform == "linux":

            # 0/ Open dataset with chunks and create random point locations to interpolate
            ds = xr.open_dataset(fn, chunks={"x": chunksizes_in_mem[0], "y": chunksizes_in_mem[1]})
            darr = ds["test"].data

            rng = np.random.default_rng(seed=42)
            interp_x = rng.choice(ds.x.size, ninterp) + rng.random(ninterp)
            interp_y = rng.choice(ds.y.size, ninterp) + rng.random(ninterp)

            # 1/ Estimation of theoretical memory usage of the subsampling script
            max_op_memusage = _estimate_interp_points_memusage(
                darr=darr, chunksizes_in_mem=chunksizes_in_mem, ninterp=ninterp
            )

            # 2/ Run interpolation of random point coordinates with memory monitoring
            interp1, measured_op_memusage = _run_dask_measuring_memusage(
                cluster, delayed_interp_points, darr, points=(interp_x, interp_y), resolution=(1, 1)
            )
            # Check the measured memory usage is smaller than the maximum estimated one
            assert measured_op_memusage < max_op_memusage

    @pytest.mark.parametrize("fn", [fn_large])  # type: ignore
    @pytest.mark.parametrize("chunksizes_in_mem", [(1000, 1000), (2500, 2500)])  # type: ignore
    @pytest.mark.parametrize("dst_chunksizes", [(1000, 1000), (2500, 2500)])  # type: ignore
    @pytest.mark.parametrize("dst_bounds_rel_shift", [(1, 1), (2, 2)])  # type: ignore
    def test_delayed_reproject__memusage(
        self,
        fn: str,
        chunksizes_in_mem: tuple[int, int],
        dst_chunksizes: tuple[int, int],
        dst_bounds_rel_shift: tuple[float, float],
        cluster: Any,
    ):
        """
        Checks for the delayed reproject function for memory usage on a big file.
        Variables that influence memory usage are:
        - Source chunksizes in memory,
        - Destination chunksizes in memory,
        - Relative difference in resolution (potentially more/less source chunks to load for a destination chunk).
        """

        # Only check on linux
        if sys.platform == "linux":

            # We fix arbitrary changes to the destination shape/resolution/bounds
            # (already checked in details in the output tests)
            dst_shape_diff = (25, -25)
            dst_res_rel_fac = (1.5, 0.5)

            # 0/ Open dataset with chunks and define variables
            ds = xr.open_dataset(fn, chunks={"x": chunksizes_in_mem[0], "y": chunksizes_in_mem[1]})
            darr = ds["test"].data

            # Get input and output shape
            src_shape = darr.shape
            dst_shape = (src_shape[0], src_shape[1] + dst_shape_diff[1])

            # Define arbitrary input/output CRS, they don't have a direct influence on the delayed method
            # (as long as the input/output transforms intersect if projected in the same CRS)
            src_crs = CRS(4326)
            dst_crs = CRS(32630)

            # Define arbitrary input transform, as we only care about the relative difference with the output transform
            src_transform = rio.transform.from_bounds(10, 10, 15, 15, src_shape[0], src_shape[1])

            # Other arguments having no influence
            src_nodata = -9999
            dst_nodata = 99999
            resampling = rio.enums.Resampling.bilinear

            # Get shifted dst_transform with new resolution
            dst_transform = _build_dst_transform_shifted_newres(
                src_transform=src_transform,
                src_crs=src_crs,
                dst_crs=dst_crs,
                src_shape=src_shape,
                bounds_rel_shift=dst_bounds_rel_shift,
                res_rel_fac=dst_res_rel_fac,
            )

            # 1/ Estimation of theoretical memory usage of the subsampling script

            max_op_memusage = _estimate_reproject_memusage(
                darr, chunksizes_in_mem=chunksizes_in_mem, dst_chunksizes=dst_chunksizes, rel_res_fac=dst_res_rel_fac
            )

            # 2/ Run delayed reproject with memory monitoring

            # We define a function where computes happens during writing to be able to measure memory usage
            # (delayed_reproject returns a delayed array that might not fit in memory, unlike subsampling/interpolation)
            fn_tmp_out = os.path.join(_EXAMPLES_DIRECTORY, os.path.splitext(os.path.basename(fn))[0] + "_reproj.nc")

            def reproject_and_write(*args: Any, **kwargs: Any) -> None:
                # Run delayed reprojection
                reproj_arr_tmp = delayed_reproject(*args, **kwargs)

                # Save file out-of-memory and compute
                data_arr = xr.DataArray(data=reproj_arr_tmp, dims=["x", "y"])
                ds_out = xr.Dataset(data_vars={"test_reproj": data_arr})
                write_delayed = ds_out.to_netcdf(fn_tmp_out, compute=False)
                write_delayed.compute()

            # And call this function with memory usage monitoring
            _, measured_op_memusage = _run_dask_measuring_memusage(
                cluster,
                reproject_and_write,
                darr,
                src_transform=src_transform,
                src_crs=src_crs,
                dst_transform=dst_transform,
                dst_crs=dst_crs,
                dst_shape=dst_shape,
                src_nodata=src_nodata,
                dst_nodata=dst_nodata,
                resampling=resampling,
                dst_chunksizes=dst_chunksizes,
            )

            # Check the measured memory usage is smaller than the maximum estimated one
            assert measured_op_memusage < max_op_memusage



# Copyright (c) 2025 GeoUtils developers
#
# This file is part of the GeoUtils project:
# https://github.com/glaciohack/geoutils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Multiple rasters tools."""

from __future__ import annotations

import warnings
from typing import Any, Callable

import numpy as np
import rasterio as rio
import rasterio.warp
from tqdm import tqdm

import geoutils as gu
from geoutils._typing import NDArrayNum
from geoutils.raster.array import get_array_and_mask
from geoutils.raster.geotransformations import _resampling_method_from_str
from geoutils.raster.raster import RasterType, _default_nodata


def load_multiple_rasters(
    raster_paths: list[str], crop: bool = True, ref_grid: int | None = None, **kwargs: Any
) -> list[RasterType]:
    """
    Function to load multiple rasters at once in a memory efficient way.
    First load metadata only.
    Optionally, crop all rasters to their intersection (default).
    Optionally, reproject all rasters to the grid of one raster set as reference (after optional crop).
    Otherwise, simply load the full rasters.

    :param raster_paths: List of paths to the rasters to be loaded.
    :param crop: If set to True, will only load rasters in the area they intersect.
    :param ref_grid: If set to an integer value, the raster with that index will be considered as the reference
        and all other rasters will be reprojected on the same grid (after optional crop).
    :param kwargs: Optional arguments to be passed to Raster.reproject, e.g. the resampling method.

    :returns: List of loaded Raster instances.
    """
    # If ref_grid is provided, need to reproject
    if isinstance(ref_grid, int):
        reproject = True
    # if no ref_grid provided, still need a reference CRS, use first by default
    elif ref_grid is None:
        ref_grid = 0
        reproject = False
    else:
        raise ValueError("`ref_grid` must be None or an integer")

    # Need to define a reference CRS for calculating intersection
    ref_crs = gu.Raster(raster_paths[ref_grid], load_data=False).crs

    # First load all rasters metadata
    output_rst = []
    bounds = []
    for path in raster_paths:
        # Initialize raster
        rst = gu.Raster(path, load_data=False)
        output_rst.append(rst)

        # Get bound in reference CRS
        bound = rst.get_bounds_projected(ref_crs)
        bounds.append(bound)

    # Second get the intersection of all raster bounds
    intersection = gu.projtools.merge_bounds(bounds, merging_algorithm="intersection")

    # Optionally, crop the rasters
    if crop:
        # Check that intersection is not void (changed to NaN instead of empty tuple end 2022)
        if intersection == () or all(np.isnan(i) for i in intersection):
            warnings.warn("Intersection is void, returning unloaded rasters.")
            return output_rst

        for rst in output_rst:
            # Calculate bounds in rst's CRS
            # rasterio's default for densify_pts is too low for very large images, set a default of 5000
            new_bounds = rio.warp.transform_bounds(
                ref_crs, rst.crs, intersection[0], intersection[1], intersection[2], intersection[3], densify_pts=5000
            )
            # Ensure bounds align with the original ones, to avoid resampling at this stage
            new_bounds = gu.projtools.align_bounds(rst.transform, new_bounds)
            rst.crop(new_bounds, mode="match_pixel", inplace=True)

    # Optionally, reproject all rasters to the reference grid
    if reproject:
        ref_rst = output_rst[ref_grid]

        # Set output bounds - intersection if crop is True, otherwise use that of ref_grid
        if crop:
            # make sure new bounds align with reference's bounds (to avoid resampling ref)
            new_bounds = intersection
            new_bounds = gu.projtools.align_bounds(ref_rst.transform, intersection)
            new_bounds = {"left": new_bounds[0], "bottom": new_bounds[1], "right": new_bounds[2], "top": new_bounds[3]}
        else:
            new_bounds = ref_rst.bounds

        # Reproject all rasters
        for index, rst in enumerate(output_rst):
            out_rst = rst.reproject(crs=ref_rst.crs, bounds=new_bounds, res=ref_rst.res, silent=True, **kwargs)
            if not out_rst.is_loaded:
                out_rst.load()
            output_rst[index] = out_rst

    # if no crop or reproject option, simply load the rasters
    if (not crop) & (not reproject):
        for rst in output_rst:
            rst.load()

    return output_rst


def stack_rasters(
    rasters: list[RasterType],
    reference: int | gu.Raster = 0,
    resampling_method: str | rio.enums.Resampling = "bilinear",
    use_ref_bounds: bool = False,
    diff: bool = False,
    progress: bool = True,
) -> gu.Raster:
    """
    Stack a list of rasters on their maximum extent into a multi-band raster.

    The input rasters can have any transform or CRS, and will be reprojected to the
    reference raster's CRS and resolution.
    The output multi-band raster has an extent that is the union of all raster extents,
    except if `use_ref_bounds` is used,
    and the number of band equal to the number of input rasters.

    Use diff=True to return directly the difference to the reference raster.

    Note that all rasters will be loaded once in memory. The data is only loaded for
    reprojection then deleted to optimize memory usage.

    :param rasters: List of rasters to be stacked.
    :param reference: Index of reference raster in the list or separate reference raster.
        Defaults to the first raster in the list.
    :param resampling_method: Resampling method for reprojection.
    :param use_ref_bounds: If True, will use reference bounds, otherwise will use maximum bounds of all rasters.
    :param diff: If True, will return the difference to the reference raster.
    :param progress: If True, will display a progress bar. Default is True.

    :returns: The merged raster with same CRS and resolution (and optionally bounds) as the reference.
    """
    # Check resampling method
    if isinstance(resampling_method, str):
        resampling_method = _resampling_method_from_str(resampling_method)

    # Check raster has a single band
    if any(r.count > 1 for r in rasters):
        warnings.warn("Some input Rasters have multiple bands, only their first band will be used.")

    # Select reference raster
    if isinstance(reference, int):
        reference_raster = rasters[reference]
    elif isinstance(reference, gu.Raster):
        reference_raster = reference
    else:
        raise ValueError("reference should be either an integer or geoutils.Raster object")

    # Set output bounds
    if use_ref_bounds:
        dst_bounds = reference_raster.bounds
    else:
        dst_bounds = gu.projtools.merge_bounds(
            [raster.get_bounds_projected(out_crs=reference_raster.crs) for raster in rasters],
            resolution=reference_raster.res[0],
            return_rio_bbox=True,
        )

    # Make a data list and add all the reprojected rasters into it.
    data: list[NDArrayNum] = []

    for raster in tqdm(rasters, disable=not progress):
        # Check that data is loaded, otherwise temporarily load it
        if not raster.is_loaded:
            raster.load()

        nodata = reference_raster.nodata if not None else gu.raster.raster._default_nodata(reference_raster.data.dtype)
        # Reproject to reference grid
        reprojected_raster = raster.reproject(
            bounds=dst_bounds,
            res=reference_raster.res,
            crs=reference_raster.crs,
            dtype=reference_raster.data.dtype,
            nodata=nodata,
            resampling=resampling_method,
            silent=True,
        )
        # If the georeferenced grid was the same, reproject() will have returned self with a warning (silenced here),
        # and we want to copy the raster and just modify its nodata (or would modify raster inputs of this function)
        if reprojected_raster.georeferenced_grid_equal(raster):
            reprojected_raster = reprojected_raster.copy()
            reprojected_raster.set_nodata(nodata)

        # Optionally calculate difference
        if diff:
            diff_to_ref = (reference_raster.data - reprojected_raster.data).squeeze()
            diff_to_ref, _ = get_array_and_mask(diff_to_ref)
            data.append(diff_to_ref)
        else:
            # img_data, _ = get_array_and_mask(reprojected_raster.data.squeeze())
            # Use only first band
            if reprojected_raster.count == 1:
                data.append(reprojected_raster.data[:])
            else:
                data.append(reprojected_raster.data[0, :])

        # Remove unloaded rasters
        if not raster.is_loaded:
            raster._data = None

    # Convert to masked array
    data = np.ma.asarray(data)
    if reference_raster.nodata is not None:
        nodata = reference_raster.nodata
    else:
        nodata = _default_nodata(data.dtype)
    data[np.isnan(data)] = nodata  # type: ignore

    # Save as gu.Raster - needed as some child classes may not accept multiple bands
    r = gu.Raster.from_array(
        data=data,
        transform=rio.transform.from_bounds(*dst_bounds, width=data[0].shape[1], height=data[0].shape[0]),
        crs=reference_raster.crs,
        nodata=nodata,
    )

    return r


def merge_rasters(
    rasters: list[RasterType],
    reference: int | RasterType = 0,
    merge_algorithm: Callable | list[Callable] = np.nanmean,  # type: ignore
    resampling_method: str | rio.enums.Resampling = "bilinear",
    use_ref_bounds: bool = False,
    progress: bool = True,
) -> RasterType:
    """
    Spatially merge a list of rasters into one larger raster of their maximum extent.

    The input rasters can have any transform or CRS, and will be reprojected to the
    reference raster's CRS and resolution.
    The output merged raster has an extent that is the union of all raster extents,
    except if `use_ref_bounds` is used.

    Note that all rasters will be loaded once in memory. The data is only loaded for
    reprojection then deleted to optimize memory usage.

    :param rasters: List of rasters to be merged.
    :param reference: Index of reference raster in the list or separate reference raster.
        Defaults to the first raster in the list.
    :param merge_algorithm: Reductor function (or list of functions) to merge the rasters with. Defaults to the mean.
        If several algorithms are provided, each result is returned as a separate band.
    :param resampling_method: Resampling method for reprojection.
    :param use_ref_bounds: If True, will use reference bounds, otherwise will use maximum bounds of all rasters.
    :param progress: If True, will display a progress bar. Default is True.

    :returns: The merged raster with same CRS and resolution (and optionally bounds) as the reference.
    """
    # Make sure merge_algorithm is a list
    if not isinstance(merge_algorithm, (list, tuple)):
        merge_algorithm = [
            merge_algorithm,
        ]

    # Try to run the merge_algorithm with an arbitrary list. Raise an error if the algorithm is incompatible.
    for algo in merge_algorithm:
        try:
            algo([1, 2])
        except TypeError as exception:
            raise TypeError(f"merge_algorithm must be able to take a list as its first argument.\n\n{exception}")

    # Select reference raster
    if isinstance(reference, int):
        reference_raster = rasters[reference]
    elif isinstance(reference, gu.Raster):
        reference_raster = reference
    else:
        raise ValueError("reference should be either an integer or geoutils.Raster object")

    # Reproject and stack all rasters
    raster_stack = stack_rasters(
        rasters,
        reference=reference,
        resampling_method=resampling_method,
        use_ref_bounds=use_ref_bounds,
        progress=progress,
    )

    # Try to use the keyword axis=0 for the merging algorithm (if it's a numpy ufunc).
    merged_data = []
    for algo in merge_algorithm:
        try:
            merged_data.append(algo(raster_stack.data, axis=0))
        # If that doesn't work, use the slower np.apply_along_axis approach.
        except TypeError as exception:
            if "'axis' is an invalid keyword" not in str(exception):
                raise exception
            merged_data.append(np.apply_along_axis(algo, axis=0, arr=raster_stack.data))

    # Convert to masked array, and set all Nans to nodata
    merged_data = np.ma.asarray(merged_data)
    if reference_raster.nodata is not None:
        nodata = reference_raster.nodata
    else:
        nodata = _default_nodata(merged_data.dtype)
    merged_data[np.isnan(merged_data)] = nodata

    # Save as gu.Raster
    merged_raster = reference_raster.from_array(
        data=np.reshape(merged_data, (len(merged_data),) + merged_data[0].shape),
        transform=rio.transform.from_bounds(
            *raster_stack.bounds, width=merged_data[0].shape[1], height=merged_data[0].shape[0]
        ),
        crs=reference_raster.crs,
        nodata=nodata,
    )

    return merged_raster



# Copyright (c) 2025 GeoUtils developers
#
# This file is part of the GeoUtils project:
# https://github.com/glaciohack/geoutils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functionalities for geotransformations of vectors."""

from __future__ import annotations

import geopandas as gpd
import rasterio as rio
from rasterio.crs import CRS

import geoutils as gu


def _reproject(
    gdf: gpd.GeoDataFrame,
    ref: gu.Raster | rio.io.DatasetReader | gu.Vector | gpd.GeoDataFrame | None = None,
    crs: CRS | str | int | None = None,
) -> gpd.GeoDataFrame:
    """Reproject a vector. See Vector.reproject() for more details."""

    # Check that either ref or crs is provided
    if (ref is not None and crs is not None) or (ref is None and crs is None):
        raise ValueError("Either of `ref` or `crs` must be set. Not both.")

    # Case a raster or vector is provided as reference
    if ref is not None:
        # Check that ref type is either str, Raster or rasterio data set
        if isinstance(ref, (gu.Raster, gu.Vector, rio.io.DatasetReader, gpd.GeoDataFrame)):
            ds_ref = ref
        else:
            raise TypeError("Type of ref must be a raster or vector.")

        # Read reprojecting params from ref raster
        crs = ds_ref.crs
    else:
        # Determine user-input target CRS
        crs = CRS.from_user_input(crs)

    new_ds = gdf.to_crs(crs=crs)

    return new_ds



# Copyright (c) 2025 GeoUtils developers
#
# This file is part of the GeoUtils project:
# https://github.com/glaciohack/geoutils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tiling tools for arrays and rasters."""

import math

import numpy as np
from matplotlib.patches import Rectangle

from geoutils._typing import NDArrayNum
from geoutils.raster import RasterType


def _get_closest_rectangle(size: int) -> tuple[int, int]:
    """
    Given a 1D array size, return a rectangular shape that is closest to a cube which the size fits in.

    If 'size' does not have an integer root, a rectangle is returned that is slightly larger than 'size'.

    :examples:
        >>> _get_closest_rectangle(4)  # size will be 4
        (2, 2)
        >>> _get_closest_rectangle(9)  # size will be 9
        (3, 3)
        >>> _get_closest_rectangle(3)  # size will be 4; needs padding afterward.
        (2, 2)
        >>> _get_closest_rectangle(55) # size will be 56; needs padding afterward.
        (7, 8)
        >>> _get_closest_rectangle(24)  # size will be 25; needs padding afterward
        (5, 5)
        >>> _get_closest_rectangle(85620)  # size will be 85849; needs padding afterward
        (293, 293)
        >>> _get_closest_rectangle(52011)  # size will be 52212; needs padding afterward
        (228, 229)
    """
    close_cube = int(np.sqrt(size))

    # If size has an integer root, return the respective cube.
    if close_cube**2 == size:
        return close_cube, close_cube

    # One of these rectangles/cubes will cover all cells, so return the first that does.
    potential_rectangles = [(close_cube, close_cube + 1), (close_cube + 1, close_cube + 1)]

    for rectangle in potential_rectangles:
        if np.prod(rectangle) >= size:
            return rectangle

    # Default return, should never reach here as one of the potential_rectangles will cover all cells.
    return potential_rectangles[-1]


def subdivide_array(shape: tuple[int, ...], count: int) -> NDArrayNum:
    """
    Create indices for subdivison of an array in a number of blocks.

    If 'count' is divisible by the product of 'shape', the amount of cells in each block will be equal.
    If 'count' is not divisible, the amount of cells in each block will be very close to equal.

    :param shape: The shape of a array to be subdivided.
    :param count: The amount of subdivisions to make.

    :examples:
        >>> subdivide_array((4, 4), 4)
        array([[0, 0, 1, 1],
               [0, 0, 1, 1],
               [2, 2, 3, 3],
               [2, 2, 3, 3]])

        >>> subdivide_array((6, 4), 4)
        array([[0, 0, 1, 1],
               [0, 0, 1, 1],
               [0, 0, 1, 1],
               [2, 2, 3, 3],
               [2, 2, 3, 3],
               [2, 2, 3, 3]])

        >>> subdivide_array((5, 4), 3)
        array([[0, 0, 0, 0],
               [0, 0, 0, 0],
               [1, 1, 2, 2],
               [1, 1, 2, 2],
               [1, 1, 2, 2]])

    :raises ValueError: If the 'shape' size (`np.prod(shape)`) is smallern than 'count'
                        If the shape is not a 2D shape.

    :returns: An array of shape 'shape' with 'count' unique indices.
    """
    try:
        import skimage.transform
    except ImportError:
        raise ImportError("Missing optional dependency, skimage.transform, required by this function.")

    if count > np.prod(shape):
        raise ValueError(f"Shape '{shape}' size ({np.prod(shape)}) is smaller than 'count' ({count}).")

    if len(shape) != 2:
        raise ValueError(f"Expected a 2D shape, got {len(shape)}D shape: {shape}")

    # Generate a small grid of indices, with the same unique count as 'count'
    rect = _get_closest_rectangle(count)
    small_indices = np.pad(np.arange(count), np.prod(rect) - count, mode="edge")[: int(np.prod(rect))].reshape(rect)

    # Upscale the grid to fit the output shape using nearest neighbour scaling.
    indices = skimage.transform.resize(small_indices, shape, order=0, preserve_range=True).astype(int)

    return indices.reshape(shape)


def _generate_tiling_grid(
    row_min: int,
    col_min: int,
    row_max: int,
    col_max: int,
    row_split: int,
    col_split: int,
    overlap: int = 0,
) -> NDArrayNum:
    """
    Generate a grid of positions by splitting [row_min, row_max] x
    [col_min, col_max] into tiles of size row_split x col_split with optional overlap.

    :param row_min: Minimum row index of the bounding box to split.
    :param col_min: Minimum column index of the bounding box to split.
    :param row_max: Maximum row index of the bounding box to split.
    :param col_max: Maximum column index of the bounding box to split.
    :param row_split: Height of each tile.
    :param col_split: Width of each tile.
    :param overlap: size of overlapping between tiles (both vertically and horizontally).
    :return: A numpy array grid with splits in two dimensions (0: row, 1: column),
             where each cell contains [row_min, row_max, col_min, col_max].
    """
    # Calculate the number of splits considering overlap
    nb_col_split = math.ceil((col_max - col_min) / col_split)
    nb_row_split = math.ceil((row_max - row_min) / row_split)

    # Initialize the output grid
    tiling_grid = np.zeros(shape=(nb_row_split, nb_col_split, 4), dtype=int)

    for row in range(nb_row_split):
        for col in range(nb_col_split):
            # Calculate the start of the tile
            row_start = max(row_min + row * row_split - overlap, 0)
            col_start = max(col_min + col * col_split - overlap, 0)

            # Calculate the end of the tile ensuring it doesn't exceed the bounds
            row_end = min(row_max, (row + 1) * row_split + overlap)
            col_end = min(col_max, (col + 1) * col_split + overlap)

            # Populate the grid with the tile boundaries
            tiling_grid[row, col] = [row_start, row_end, col_start, col_end]

    return tiling_grid


def compute_tiling(
    tile_size: int,
    raster_shape: tuple[int, int],
    ref_shape: tuple[int, int],
    overlap: int = 0,
) -> NDArrayNum:
    """
    Compute the raster tiling grid to coregister raster by block.

    :param tile_size: Size of each tile (square tiles).
    :param raster_shape: Shape of the raster to determine tiling parameters.
    :param ref_shape: The shape of another raster to coregister, use to validate the shape.
    :param overlap: Size of overlap between tiles (optional).
    :return: tiling_grid (array of tile boundaries).
    """
    if raster_shape != ref_shape:
        raise Exception("Reference and secondary rasters do not have the same shape")
    row_max, col_max = raster_shape

    # Generate tiling
    tiling_grid = _generate_tiling_grid(0, 0, row_max, col_max, tile_size, tile_size, overlap=overlap)
    return tiling_grid


def plot_tiling(raster: RasterType, tiling_grid: NDArrayNum) -> None:
    """
    Plot raster with its tiling.

    :param raster: The raster to plot with its tiling.
    :param tiling_grid: tiling given by compute_tiling.
    """
    ax, caxes = raster.plot(return_axes=True)
    for tile in tiling_grid.reshape(-1, 4):
        row_min, row_max, col_min, col_max = tile
        x_min, y_min = raster.transform * (col_min, row_min)  # Bottom-left corner
        x_max, y_max = raster.transform * (col_max, row_max)  # Top-right corne
        rect = Rectangle((x_min, y_min), x_max - x_min, y_max - y_min, edgecolor="red", facecolor="none", linewidth=1.5)
        ax.add_patch(rect)



"""
Polygonize a raster
===================

This example demonstrates the polygonizing of a raster using :func:`geoutils.Raster.polygonize` and :func:`geoutils.Mask.polygonize`.
"""

# %%
# We open a raster.

# sphinx_gallery_thumbnail_number = 3
import geoutils as gu

filename_rast = gu.examples.get_path("exploradores_aster_dem")
rast = gu.Raster(filename_rast)
rast = rast.crop([rast.bounds.left, rast.bounds.bottom, rast.bounds.left + 5000, rast.bounds.bottom + 5000])
# %%
# Let's plot the raster.
rast.plot(cmap="terrain")

# %%
# We polygonize the raster.

rast_polygonized = rast.polygonize()
rast_polygonized.plot(ax="new")

# %%
# By default, :func:`~geoutils.Raster.polygonize` will try to polygonize target all valid values. Instead, one can specify discrete values to target by
# passing a number or :class:`list`, or a range of values by passing a :class:`tuple`.

# A range of values to polygonize
rast_polygonized = rast.polygonize((2500, 3000))
rast_polygonized.plot(ax="new")

# %%
# An even simpler way to do this is to compute a :func:`~geoutils.Mask` to polygonize using logical comparisons on the :func:`~geoutils.Raster`.

rast_polygonized = ((2500 < rast) & (rast < 3000)).polygonize()
rast_polygonized.plot(ax="new")

# %%
# .. note::
#           See :ref:`core-py-ops` for more details on casting to :func:`~geoutils.Mask`.



# Copyright (c) 2025 GeoUtils developers
#
# This file is part of the GeoUtils project:
# https://github.com/glaciohack/geoutils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functionalities to manipulate vector geometries."""

from __future__ import annotations

import warnings

import geopandas as gpd
import matplotlib.pyplot as plt
import numpy as np
import shapely
from scipy.spatial import Voronoi
from shapely.geometry.polygon import Polygon

import geoutils as gu
from geoutils.projtools import _get_utm_ups_crs, bounds2poly


def _buffer_metric(gdf: gpd.GeoDataFrame, buffer_size: float) -> gu.Vector:
    """
    Metric buffering. See Vector.buffer_metric() for details.
    """

    crs_utm_ups = _get_utm_ups_crs(df=gdf)

    # Reproject the shapefile in the local UTM
    ds_utm = gdf.to_crs(crs=crs_utm_ups)

    # Buffer the shapefile
    ds_buffered = ds_utm.buffer(distance=buffer_size)
    del ds_utm

    # Revert-project the shapefile in the original CRS
    ds_buffered_origproj = ds_buffered.to_crs(crs=gdf.crs)
    del ds_buffered

    # Return a Vector object of the buffered GeoDataFrame
    # TODO: Clarify what is conserved in the GeoSeries and what to pass the GeoDataFrame to not lose any attributes
    vector_buffered = gu.Vector(gpd.GeoDataFrame(geometry=ds_buffered_origproj.geometry, crs=gdf.crs))

    return vector_buffered


def _buffer_without_overlap(
    ds: gpd.GeoDataFrame, buffer_size: int | float, metric: bool = True, plot: bool = False
) -> gu.Vector:
    """See Vector.buffer_without_overlap() for details."""

    # Project in local UTM if metric is True
    if metric:
        crs_utm_ups = _get_utm_ups_crs(df=ds)
        gdf = ds.to_crs(crs=crs_utm_ups)
    else:
        gdf = ds

    # Dissolve all geometries into one
    merged = gdf.dissolve()

    # Add buffer around geometries
    merged_buffer = merged.buffer(buffer_size)

    # Extract only the buffered area
    buffer = merged_buffer.difference(merged)

    # Crop Voronoi polygons to bound geometry and add missing polygons
    bound_poly = bounds2poly(gdf)
    bound_poly = bound_poly.buffer(buffer_size)
    voronoi_all = _generate_voronoi_with_bounds(gdf, bound_poly)
    if plot:
        plt.figure(figsize=(16, 4))
        ax1 = plt.subplot(141)
        voronoi_all.plot(ax=ax1)
        gdf.plot(fc="none", ec="k", ax=ax1)
        ax1.set_title("Voronoi polygons, cropped")

    # Extract Voronoi polygons only within the buffer area
    voronoi_diff = voronoi_all.intersection(buffer.geometry[0])

    # Split all polygons, and join attributes of original geometries into the Voronoi polygons
    # Splitting, i.e. explode, is needed when Voronoi generate MultiPolygons that may extend over several features.
    voronoi_gdf = gpd.GeoDataFrame(geometry=voronoi_diff.explode(index_parts=True))  # requires geopandas>=0.10
    joined_voronoi = gpd.tools.sjoin(gdf, voronoi_gdf, how="right")

    # Plot results -> some polygons are duplicated
    if plot:
        ax2 = plt.subplot(142, sharex=ax1, sharey=ax1)
        joined_voronoi.plot(ax=ax2, column="index_left", alpha=0.5, ec="k")
        gdf.plot(ax=ax2, column=gdf.index.values)
        ax2.set_title("Buffer with duplicated polygons")

    # Find non unique Voronoi polygons, and retain only first one
    _, indexes = np.unique(joined_voronoi.index, return_index=True)
    unique_voronoi = joined_voronoi.iloc[indexes]

    # Plot results -> unique polygons only
    if plot:
        ax3 = plt.subplot(143, sharex=ax1, sharey=ax1)
        unique_voronoi.plot(ax=ax3, column="index_left", alpha=0.5, ec="k")
        gdf.plot(ax=ax3, column=gdf.index.values)
        ax3.set_title("Buffer with unique polygons")

    # Dissolve all polygons by original index
    merged_voronoi = unique_voronoi.dissolve(by="index_left")

    # Plot
    if plot:
        ax4 = plt.subplot(144, sharex=ax1, sharey=ax1)
        gdf.plot(ax=ax4, column=gdf.index.values)
        merged_voronoi.plot(column=merged_voronoi.index.values, ax=ax4, alpha=0.5)
        ax4.set_title("Final buffer")
        plt.show()

    # Reverse-project to the original CRS if metric is True
    if metric:
        merged_voronoi = merged_voronoi.to_crs(crs=ds.crs)

    return gu.Vector(merged_voronoi)


def _extract_vertices(gdf: gpd.GeoDataFrame) -> list[list[tuple[float, float]]]:
    r"""
    Function to extract the exterior vertices of all shapes within a gpd.GeoDataFrame.

    :param gdf: The GeoDataFrame from which the vertices need to be extracted.

    :returns: A list containing a list of (x, y) positions of the vertices. The length of the primary list is equal
        to the number of geometries inside gdf, and length of each sublist is the number of vertices in the geometry.
    """
    vertices = []
    # Loop on all geometries within gdf
    for geom in gdf.geometry:
        # Extract geometry exterior(s)
        if geom.geom_type == "MultiPolygon":
            exteriors = [p.exterior for p in geom.geoms]
        elif geom.geom_type == "Polygon":
            exteriors = [geom.exterior]
        elif geom.geom_type == "LineString":
            exteriors = [geom]
        elif geom.geom_type == "MultiLineString":
            exteriors = list(geom.geoms)
        else:
            raise NotImplementedError(f"Geometry type {geom.geom_type} not implemented.")

        vertices.extend([list(ext.coords) for ext in exteriors])

    return vertices


def _generate_voronoi_polygons(gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
    """
    Generate Voronoi polygons (tessellation) from the vertices of all geometries in a GeoDataFrame.

    Uses scipy.spatial.voronoi.

    :param: The GeoDataFrame from whose vertices are used for the Voronoi polygons.

    :returns: A GeoDataFrame containing the Voronoi polygons.
    """
    # Extract the coordinates of the vertices of all geometries in gdf
    vertices = _extract_vertices(gdf)
    coords = np.concatenate(vertices)

    # Create the Voronoi diagram and extract ridges
    vor = Voronoi(coords)
    lines = [shapely.geometry.LineString(vor.vertices[line]) for line in vor.ridge_vertices if -1 not in line]
    polys = list(shapely.ops.polygonize(lines))
    if len(polys) == 0:
        raise ValueError("Invalid geometry, cannot generate finite Voronoi polygons")

    # Convert into GeoDataFrame
    voronoi = gpd.GeoDataFrame(geometry=gpd.GeoSeries(polys))
    voronoi.crs = gdf.crs

    return voronoi


def _generate_voronoi_with_bounds(gdf: gpd.GeoDataFrame, bound_poly: Polygon) -> gpd.GeoDataFrame:
    """
    Generate Voronoi polygons that are bounded by the polygon bound_poly, to avoid Voronoi polygons that extend \
far beyond the original geometry.

    Voronoi polygons are created using generate_voronoi_polygons, cropped to the extent of bound_poly and gaps \
are filled with new polygons.

    :param: The GeoDataFrame from whose vertices are used for the Voronoi polygons.
    :param: A shapely Polygon to be used for bounding the Voronoi diagrams.

    :returns: A GeoDataFrame containing the Voronoi polygons.
    """
    # Create Voronoi polygons
    voronoi = _generate_voronoi_polygons(gdf)

    # Crop Voronoi polygons to input bound_poly extent
    voronoi_crop = voronoi.intersection(bound_poly)
    voronoi_crop = gpd.GeoDataFrame(geometry=voronoi_crop)  # convert to DataFrame

    # Dissolve all Voronoi polygons and subtract from bounds to get gaps
    voronoi_merged = voronoi_crop.dissolve()
    bound_gdf = gpd.GeoDataFrame(geometry=gpd.GeoSeries(bound_poly))
    bound_gdf.crs = gdf.crs
    gaps = bound_gdf.difference(voronoi_merged)

    # Merge cropped Voronoi with gaps, if not empty, otherwise return cropped Voronoi
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", "Geometry is in a geographic CRS. Results from 'area' are likely incorrect.")
        tot_area = np.sum(gaps.area.values)

    if not tot_area == 0:
        voronoi_all = gpd.GeoDataFrame(geometry=list(voronoi_crop.geometry) + list(gaps.geometry))
        voronoi_all.crs = gdf.crs
        return voronoi_all
    else:
        return voronoi_crop



"""
Open/save a vector
==================

This example demonstrates the instantiation of a vector through :class:`geoutils.Vector` and saving with :func:`~geoutils.Vector.save`.
"""

import geoutils as gu

# %%
# We open an example vector.
filename_vect = gu.examples.get_path("everest_rgi_outlines")
vect = gu.Vector(filename_vect)
vect

# %%
# A vector is composed of a single main attribute: a :class:`~geoutils.Vector.ds` geodataframe.
# All other attributes are :ref:`inherited from Shapely and GeoPandas<vector-from-geopandas>`. See also :ref:`vector-class`.

# %%
#
# .. note::
#        A vector can also be instantiated with a :class:`geopandas.GeoDataFrame`, see :ref:`sphx_glr_io_examples_import_export_import_vector.py`.
#
# We can print more info on the vector.
vect.info()

# %%
# Let's plot by vector area
vect.plot(column="Area", cbar_title="Area (km²)")

# %%
# Finally, a vector is saved using :func:`~geoutils.Vector.save`.

vect.save("myvector.gpkg")



"""Test functions specific to the Vector class."""

from __future__ import annotations

import inspect
import os.path
import pathlib
import tempfile
import warnings

import geopandas as gpd
import geopandas.base
import pyproj
import pytest
from geopandas.testing import assert_geodataframe_equal, assert_geoseries_equal
from pandas.testing import assert_series_equal
from shapely.geometry.base import BaseGeometry
from shapely.geometry.linestring import LineString
from shapely.geometry.polygon import Polygon

import geoutils as gu

GLACIER_OUTLINES_URL = "http://public.data.npolar.no/cryoclim/CryoClim_GAO_SJ_1990.zip"


class TestVector:
    landsat_b4_crop_path = gu.examples.get_path("everest_landsat_b4_cropped")
    everest_outlines_path = gu.examples.get_path("everest_rgi_outlines")
    aster_dem_path = gu.examples.get_path("exploradores_aster_dem")
    aster_outlines_path = gu.examples.get_path("exploradores_rgi_outlines")
    glacier_outlines = gu.Vector(GLACIER_OUTLINES_URL)

    def test_init(self) -> None:
        """Test class initiation works as intended"""

        # First, with a URL filename
        v = gu.Vector(GLACIER_OUTLINES_URL)
        assert isinstance(v, gu.Vector)

        # Second, with a string filename
        v0 = gu.Vector(self.aster_outlines_path)
        assert isinstance(v0, gu.Vector)

        # Third, with a pathlib path
        path = pathlib.Path(self.aster_outlines_path)
        v1 = gu.Vector(path)
        assert isinstance(v1, gu.Vector)

        # Fourth, with a geopandas dataframe
        v2 = gu.Vector(gpd.read_file(self.aster_outlines_path))
        assert isinstance(v2, gu.Vector)

        # Fifth, passing a Vector itself (points back to Vector passed)
        v3 = gu.Vector(v2)
        assert isinstance(v3, gu.Vector)

        # Check errors are raised when filename has wrong type
        with pytest.raises(TypeError, match="Filename argument should be a string, Path or geopandas.GeoDataFrame."):
            gu.Vector(1)  # type: ignore

    def test_copy(self) -> None:
        vector2 = self.glacier_outlines.copy()

        assert vector2 is not self.glacier_outlines

        vector2.ds = vector2.ds.query("NAME == 'Ayerbreen'")

        assert vector2.ds.shape[0] < self.glacier_outlines.ds.shape[0]

    def test_info(self) -> None:

        v = gu.Vector(GLACIER_OUTLINES_URL)

        # Check default runs without error (prints to screen)
        output = v.info()
        assert output is None

        # Otherwise returns info
        output2 = v.info(verbose=False)
        assert isinstance(output2, str)
        list_prints = ["Filename", "Coordinate system", "Extent", "Number of features", "Attributes"]
        assert all(p in output2 for p in list_prints)

    def test_query(self) -> None:
        vector2 = self.glacier_outlines.query("NAME == 'Ayerbreen'")

        assert vector2 is not self.glacier_outlines

        assert vector2.ds.shape[0] < self.glacier_outlines.ds.shape[0]

    def test_save(self) -> None:
        """Test the save wrapper for GeoDataFrame.to_file()."""

        vector = gu.Vector(self.aster_outlines_path)

        # Create a temporary file in a temporary directory
        temp_dir = tempfile.TemporaryDirectory()
        temp_file = os.path.join(temp_dir.name, "test.gpkg")

        # Save and check the file exists
        vector.save(temp_file)
        assert os.path.exists(temp_file)

        # Open and check the object is the same
        vector_save = gu.Vector(temp_file)
        vector_save.vector_equal(vector)

    def test_bounds(self) -> None:
        bounds = self.glacier_outlines.bounds

        assert bounds.left < bounds.right
        assert bounds.bottom < bounds.top

        assert bounds.left == self.glacier_outlines.ds.total_bounds[0]
        assert bounds.bottom == self.glacier_outlines.ds.total_bounds[1]
        assert bounds.right == self.glacier_outlines.ds.total_bounds[2]
        assert bounds.top == self.glacier_outlines.ds.total_bounds[3]

    def test_footprint(self) -> None:

        footprint = self.glacier_outlines.footprint

        assert isinstance(footprint, gu.Vector)
        assert footprint.vector_equal(self.glacier_outlines.get_footprint_projected(self.glacier_outlines.crs))


class NeedToImplementWarning(FutureWarning):
    """Warning to remember to implement new GeoPandas methods"""


class TestGeoPandasMethods:
    # Use two synthetic vectors
    poly = Polygon([(10, 10), (11, 10), (11, 11), (10, 11)])
    gdf1 = gpd.GeoDataFrame({"geometry": [poly]}, crs="EPSG:4326")
    synthvec1 = gu.Vector(gdf1)

    # Create a synthetic LineString geometry
    lines = LineString([(10, 10), (10.5, 10.5), (11, 11)])
    gdf2 = gpd.GeoDataFrame({"geometry": [lines]}, crs="EPSG:4326")
    synthvec2 = gu.Vector(gdf2)

    # Use two real-life vectors
    realvec1 = gu.Vector(gu.examples.get_path("exploradores_rgi_outlines"))
    realvec2 = gu.Vector(gu.examples.get_path("everest_rgi_outlines"))

    # Properties and methods derived from Shapely or GeoPandas
    # List of properties and methods with non-geometric output that are implemented in GeoUtils
    main_properties = ["crs", "geometry", "total_bounds"]
    nongeo_properties = [
        "area",
        "length",
        "interiors",
        "geom_type",
        "is_empty",
        "is_ring",
        "is_simple",
        "is_valid",
        "has_z",
    ]
    nongeo_methods = [
        "contains",
        "geom_equals",
        "geom_almost_equals",
        "geom_equals_exact",
        "crosses",
        "disjoint",
        "intersects",
        "overlaps",
        "touches",
        "within",
        "covers",
        "covered_by",
        "distance",
    ]

    # List of properties and methods with geometric output that are implemented in GeoUtils
    geo_properties = ["boundary", "unary_union", "centroid", "convex_hull", "envelope", "exterior"]
    geo_methods = [
        "representative_point",
        "normalize",
        "make_valid",
        "difference",
        "symmetric_difference",
        "union",
        "intersection",
        "clip_by_rect",
        "buffer",
        "simplify",
        "affine_transform",
        "translate",
        "rotate",
        "scale",
        "skew",
        "dissolve",
        "explode",
        "sjoin",
        "sjoin_nearest",
        "overlay",
        "to_crs",
        "set_crs",
        "rename_geometry",
        "set_geometry",
        "clip",
    ]
    # List of class methods
    io_methods = [
        "from_file",
        "from_postgis",
        "from_dict",
        "from_features",
        "to_feather",
        "to_parquet",
        "to_file",
        "to_postgis",
        "to_json",
        "to_wkb",
        "to_wkt",
        "to_csv",
    ]

    # List of other properties and methods
    other = ["has_sindex", "sindex", "estimate_utm_crs", "cx", "iterfeatures"]
    all_declared = (
        main_properties + nongeo_methods + nongeo_properties + geo_methods + geo_properties + other + io_methods
    )

    # Exceptions for GeoPandasBase functions not implemented (or deprecrated) in GeoSeries/GeoDataFrame
    exceptions_unimplemented = [
        "plot",
        "explore",
        "cascaded_union",
        "bounds",
        "relate",
        "project",
        "interpolate",
        "equals",
        "type",
        "convert_dtypes",
        "merge",
        "apply",
        "astype",
        "minimum_bounding_circle",
        "minimum_bounding_radius",
        "get_coordinates",
        "hilbert_distance",
        "sample_points",
        "copy",
    ]
    # Exceptions for IO/conversion that can be done directly from .ds
    all_exceptions = exceptions_unimplemented

    # Get all GeoPandasBase public methods with some exceptions
    geobase_methods = gpd.base.GeoPandasBase.__dict__.copy()

    # Get all GeoDataFrame public methods with some exceptions
    gdf_methods = gpd.GeoDataFrame.__dict__.copy()

    def test_overridden_funcs_exist(self) -> None:
        """Check that all methods listed above exist in Vector."""

        # Check that all methods declared in the class above exist in Vector
        vector_methods = gu.Vector.__dict__

        list_missing = [method for method in self.all_declared if method not in vector_methods.keys()]

        assert len(list_missing) == 0, print(f"Test method listed that is not in GeoUtils: {list_missing}")

    def test_geopandas_coverage(self) -> None:
        """Check that all existing methods of GeoPandas are overridden, with a couple exceptions."""

        # Merge the two
        all_methods = self.geobase_methods.copy()
        all_methods.update(self.gdf_methods)

        # Remove exceptions we don't want to reuse from GeoPandas (mirrored in Vector)
        name_all_methods = list(all_methods.keys())
        public_methods = [method for method in name_all_methods if method[0] != "_"]

        covered_methods = [method for method in public_methods if method not in self.all_exceptions]

        # Check that all methods declared in the class above are covered in Vector
        list_missing = [method for method in covered_methods if method not in self.all_declared]

        if len(list_missing) != 0:
            warnings.warn(
                f"New GeoPandas methods are not implemented in GeoUtils: {list_missing}", NeedToImplementWarning
            )

    @pytest.mark.parametrize("method", nongeo_methods + geo_methods)  # type: ignore
    def test_overridden_funcs_args(self, method: str) -> None:
        """Check that all methods overridden have the same arguments as in GeoPandas."""

        # Get GeoPandas class where the methods live
        if method in self.geobase_methods.keys():
            upstream_class = gpd.base.GeoPandasBase
        elif method in self.gdf_methods.keys():
            upstream_class = gpd.GeoDataFrame
        else:
            raise ValueError("Method did not belong to GeoDataFrame or GeoPandasBase class.")

        # Get a full argument inspection object for each class
        argspec_upstream = inspect.getfullargspec(getattr(upstream_class, method))
        argspec_geoutils = inspect.getfullargspec(getattr(gu.Vector, method))

        # Check that all positional arguments are the same
        if argspec_upstream.args != argspec_geoutils.args:
            warnings.warn("Argument of GeoPandas method not consistent in GeoUtils.", NeedToImplementWarning)

        # Check that the *args and **kwargs argument are declared consistently
        if argspec_upstream.varargs != argspec_geoutils.varargs:
            warnings.warn("Argument of GeoPandas method not consistent in GeoUtils.", NeedToImplementWarning)

        if argspec_upstream.varkw != argspec_geoutils.varkw:
            warnings.warn("Argument of GeoPandas method not consistent in GeoUtils.", NeedToImplementWarning)

        # Check that default argument values are the same
        if argspec_upstream.defaults != argspec_geoutils.defaults:
            warnings.warn("Default argument of GeoPandas method not consistent in GeoUtils.", NeedToImplementWarning)

    @pytest.mark.parametrize("vector", [synthvec1, synthvec2, realvec1, realvec2])  # type: ignore
    @pytest.mark.parametrize("method", nongeo_properties)  # type: ignore
    def test_nongeo_properties(self, vector: gu.Vector, method: str) -> None:
        """Check non-geometric properties are consistent with GeoPandas."""

        # Remove warnings about operations in a non-projected system, and future changes
        warnings.simplefilter("ignore", category=UserWarning)
        warnings.simplefilter("ignore", category=FutureWarning)

        # Get method for each class
        output_geoutils = getattr(vector, method)
        output_geopandas = getattr(vector.ds, method)

        # Assert equality
        assert_series_equal(output_geoutils, output_geopandas)

    @pytest.mark.parametrize("vector1", [synthvec1, realvec1])  # type: ignore
    @pytest.mark.parametrize("vector2", [synthvec2, realvec2])  # type: ignore
    @pytest.mark.parametrize("method", nongeo_methods)  # type: ignore
    def test_nongeo_methods(self, vector1: gu.Vector, vector2: gu.Vector, method: str) -> None:
        """
        Check non-geometric methods are consistent with GeoPandas.
        All these methods require two inputs ("other", "df", or "right" argument), except one.
        """

        # Remove warnings about operations in a non-projected system, and future changes
        warnings.simplefilter("ignore", category=UserWarning)
        warnings.simplefilter("ignore", category=FutureWarning)

        # Get method for each class
        if method != "geom_equals_exact":
            output_geoutils = getattr(vector1, method)(vector2)
            output_geopandas = getattr(vector1.ds, method)(vector2.ds)
        else:
            output_geoutils = getattr(vector1, method)(vector2, tolerance=0.1)
            output_geopandas = getattr(vector1.ds, method)(vector2.ds, tolerance=0.1)

        # Assert equality
        assert_series_equal(output_geoutils, output_geopandas)

    @pytest.mark.parametrize("vector", [synthvec1, synthvec2, realvec1, realvec2])  # type: ignore
    @pytest.mark.parametrize("method", geo_properties)  # type: ignore
    def test_geo_properties(self, vector: gu.Vector, method: str) -> None:
        """Check geometric properties are consistent with GeoPandas."""

        # Remove warnings about operations in a non-projected system, and future changes
        warnings.simplefilter("ignore", category=UserWarning)
        warnings.simplefilter("ignore", category=FutureWarning)

        # Get method for each class
        output_geoutils = getattr(vector, method)
        output_geopandas = getattr(vector.ds, method)

        # Assert output types
        assert isinstance(output_geoutils, gu.Vector)
        assert isinstance(output_geopandas, (gpd.GeoSeries, gpd.GeoDataFrame, BaseGeometry))

        # Separate cases depending on GeoPandas' output
        if isinstance(output_geopandas, gpd.GeoSeries):
            # Assert geoseries equality
            assert_geoseries_equal(output_geoutils.ds.geometry, output_geopandas)
        elif isinstance(output_geopandas, BaseGeometry):
            assert_geodataframe_equal(
                output_geoutils.ds, gpd.GeoDataFrame({"geometry": [output_geopandas]}, crs=vector.crs)
            )
        else:
            assert_geodataframe_equal(output_geoutils.ds, output_geopandas)

    specific_method_args = {
        "buffer": {"distance": 1},
        "clip_by_rect": {"xmin": 10.5, "ymin": 10.5, "xmax": 11, "ymax": 11},
        "affine_transform": {"matrix": [1, 1, 1, 1, 1, 1]},
        "translate": {"xoff": 1, "yoff": 1, "zoff": 0},
        "rotate": {"angle": 90},
        "scale": {"xfact": 1.1, "yfact": 1.1, "zfact": 1.1, "origin": "center"},
        "skew": {"xs": 1.1, "ys": 1.1},
        "interpolate": {"distance": 1},
        "simplify": {"tolerance": 0.1},
        "to_crs": {"crs": pyproj.CRS.from_epsg(32610)},
        "set_crs": {"crs": pyproj.CRS.from_epsg(32610), "allow_override": True},
        "rename_geometry": {"col": "lol"},
        "set_geometry": {"col": synthvec1.geometry},
        "clip": {"mask": poly},
    }

    @pytest.mark.parametrize("vector1", [synthvec1, realvec1])  # type: ignore
    @pytest.mark.parametrize("vector2", [synthvec2, realvec2])  # type: ignore
    @pytest.mark.parametrize("method", geo_methods)  # type: ignore
    def test_geo_methods(self, vector1: gu.Vector, vector2: gu.Vector, method: str) -> None:
        """Check geometric methods are consistent with GeoPandas."""

        # Remove warnings about operations in a non-projected system, and future changes
        warnings.simplefilter("ignore", category=UserWarning)
        warnings.simplefilter("ignore", category=FutureWarning)

        # Methods that require two inputs
        if method in [
            "difference",
            "symmetric_difference",
            "union",
            "intersection",
            "sjoin",
            "sjoin_nearest",
            "overlay",
        ]:
            output_geoutils = getattr(vector1, method)(vector2)
            output_geopandas = getattr(vector1.ds, method)(vector2.ds)
        # Methods that require zero input
        elif method in ["representative_point", "normalize", "make_valid", "dissolve", "explode"]:
            output_geoutils = getattr(vector1, method)()
            output_geopandas = getattr(vector1.ds, method)()
        elif method in self.specific_method_args.keys():
            output_geoutils = getattr(vector1, method)(**self.specific_method_args[method])
            output_geopandas = getattr(vector1.ds, method)(**self.specific_method_args[method])
        else:
            raise ValueError(f"The method '{method}' is not covered by this test.")

        # Assert output types
        assert isinstance(output_geoutils, gu.Vector)
        assert isinstance(output_geopandas, (gpd.GeoSeries, gpd.GeoDataFrame))

        # Separate cases depending on GeoPandas' output, and nature of the function
        # Simplify is a special case that can make geometries invalid, so adjust test
        if method == "simplify":
            # TODO: Unskip this random test failure (one index not matching) when this is fixed in GeoPandas/Shapely
            pass
            # assert_geoseries_equal(
            #     output_geopandas.make_valid(), output_geoutils.ds.geometry.make_valid(), check_less_precise=True
            # )
        # For geoseries output, check equality of it
        elif isinstance(output_geopandas, gpd.GeoSeries):
            assert_geoseries_equal(output_geoutils.ds.geometry, output_geopandas)
        # For geodataframe output, check equality
        else:
            assert_geodataframe_equal(output_geoutils.ds, output_geopandas)



"""
Test the example files used for testing and documentation
"""

import hashlib
import warnings

import pytest

import geoutils as gu
from geoutils import examples


@pytest.mark.parametrize(
    "example", ["everest_landsat_b4", "everest_landsat_b4_cropped", "everest_landsat_rgb", "exploradores_aster_dem"]
)  # type: ignore
def test_read_paths_raster(example: str) -> None:
    assert isinstance(gu.Raster(examples.get_path(example)), gu.Raster)


@pytest.mark.parametrize("example", ["everest_rgi_outlines", "exploradores_rgi_outlines"])  # type: ignore
def test_read_paths_vector(example: str) -> None:
    warnings.simplefilter("error")
    assert isinstance(gu.Vector(examples.get_path(example)), gu.Vector)


# Original sha256 obtained with `sha256sum filename`
original_sha256 = {
    "everest_landsat_b4": "271fa34e248f016f87109c8e81960caaa737558fbae110ec9e0d9e2d30d80c26",
    "everest_landsat_b4_cropped": "0e63d8e9c4770534a1ec267c91e80cd9266732184a114f0bd1aadb5a613215e6",
    "everest_landsat_rgb": "7d0505a8610fd7784cb71c03e5b242715cd1574e978c2c86553d60fd82372c30",
    "everest_rgi_outlines": "d1a5bcd4bd4731a24c2398c016a6f5a8064160fedd5bab10609adacda9ba41ef",
    "exploradores_aster_dem": "dcb0d708d042553cdd2bb4fd82c55b5674a5e0bd6ea46f1a021b396b7d300033",
    "exploradores_rgi_outlines": "19c2dac089ce57373355213fdf2fd72f601bf97f21b04c4920edb1e4384ae2b2",
}


@pytest.mark.parametrize("example", examples.available)  # type: ignore
def test_data_integrity(example: str) -> None:
    """
    Test that input data is not corrupted by checking sha265 sum
    """
    # Read file as bytes
    fbytes = open(examples.get_path(example), "rb").read()

    # Get sha256
    file_sha256 = hashlib.sha256(fbytes).hexdigest()

    assert file_sha256 == original_sha256[example]



"""Tests for raster-point interfacing."""

from __future__ import annotations

import re

import numpy as np
import pytest
import rasterio as rio

import geoutils as gu
from geoutils import examples


class TestRasterPointInterface:

    # Paths to example data
    landsat_b4_path = examples.get_path("everest_landsat_b4")
    landsat_rgb_path = examples.get_path("everest_landsat_rgb")
    aster_dem_path = examples.get_path("exploradores_aster_dem")

    def test_to_pointcloud(self) -> None:
        """Test to_pointcloud method."""

        # 1/ Single band synthetic data

        # Create a small raster to test point sampling on
        img_arr = np.arange(25, dtype="int32").reshape(5, 5)
        img0 = gu.Raster.from_array(img_arr, transform=rio.transform.from_origin(0, 5, 1, 1), crs=4326)

        # Sample the whole raster (fraction==1)
        points = img0.to_pointcloud()
        points_arr = img0.to_pointcloud(as_array=True)

        # Check output types
        assert isinstance(points, gu.Vector)
        assert isinstance(points_arr, np.ndarray)

        # Check that both outputs (array or vector) are fully consistent, order matters here
        assert np.array_equal(points.ds.geometry.x.values, points_arr[:, 0])
        assert np.array_equal(points.ds.geometry.y.values, points_arr[:, 1])
        assert np.array_equal(points.ds["b1"].values, points_arr[:, 2])

        # Validate that 25 points were sampled (equating to img1.height * img1.width) with x, y, and band0 values.
        assert points_arr.shape == (25, 3)
        assert points.ds.shape == (25, 2)  # One less column here due to geometry storing X and Y
        # Check that X, Y and Z arrays are equal to raster array input independently of value order
        x_coords, y_coords = img0.ij2xy(i=np.arange(0, 5), j=np.arange(0, 5))
        assert np.array_equal(np.sort(np.asarray(points_arr[:, 0])), np.sort(np.tile(x_coords, 5)))
        assert np.array_equal(np.sort(np.asarray(points_arr[:, 1])), np.sort(np.tile(y_coords, 5)))
        assert np.array_equal(np.sort(np.asarray(points_arr[:, 2])), np.sort(img_arr.ravel()))

        # Check that subsampling works properly
        points_arr = img0.to_pointcloud(subsample=0.2, as_array=True)
        assert points_arr.shape == (5, 3)

        # All values should be between 0 and 25
        assert all(0 <= points_arr[:, 2]) and all(points_arr[:, 2] < 25)

        # 2/ Multi-band synthetic data
        img_arr = np.arange(25, dtype="int32").reshape(5, 5)
        img_3d_arr = np.stack((img_arr, 25 + img_arr, 50 + img_arr), axis=0)
        img3d = gu.Raster.from_array(img_3d_arr, transform=rio.transform.from_origin(0, 5, 1, 1), crs=4326)

        # Sample the whole raster (fraction==1)
        points = img3d.to_pointcloud(auxiliary_data_bands=[2, 3])
        points_arr = img3d.to_pointcloud(as_array=True, auxiliary_data_bands=[2, 3])

        # Check equality between both output types
        assert np.array_equal(points.ds.geometry.x.values, points_arr[:, 0])
        assert np.array_equal(points.ds.geometry.y.values, points_arr[:, 1])
        assert np.array_equal(points.ds["b1"].values, points_arr[:, 2])
        assert np.array_equal(points.ds["b2"].values, points_arr[:, 3])
        assert np.array_equal(points.ds["b3"].values, points_arr[:, 4])

        # Check it is the right data
        assert np.array_equal(np.sort(np.asarray(points_arr[:, 0])), np.sort(np.tile(x_coords, 5)))
        assert np.array_equal(np.sort(np.asarray(points_arr[:, 1])), np.sort(np.tile(y_coords, 5)))
        assert np.array_equal(np.sort(np.asarray(points_arr[:, 2])), np.sort(img_3d_arr[0, :, :].ravel()))
        assert np.array_equal(np.sort(np.asarray(points_arr[:, 3])), np.sort(img_3d_arr[1, :, :].ravel()))
        assert np.array_equal(np.sort(np.asarray(points_arr[:, 4])), np.sort(img_3d_arr[2, :, :].ravel()))

        # With a subsample
        points_arr = img3d.to_pointcloud(as_array=True, subsample=10, auxiliary_data_bands=[2, 3])
        assert points_arr.shape == (10, 5)

        # Check the values are still good
        assert all(0 <= points_arr[:, 2]) and all(points_arr[:, 2] < 25)
        assert all(25 <= points_arr[:, 3]) and all(points_arr[:, 3] < 50)
        assert all(50 <= points_arr[:, 4]) and all(points_arr[:, 4] < 75)

        # 3/ Single-band real raster with nodata values
        img1 = gu.Raster(self.aster_dem_path)

        # Get a large sample to ensure they should be some NaNs normally
        points_arr = img1.to_pointcloud(subsample=10000, as_array=True, random_state=42)
        points = img1.to_pointcloud(subsample=10000, random_state=42)

        # This should not load the image
        assert not img1.is_loaded

        # The subsampled values should be valid and the right shape
        assert points_arr.shape == (10000, 3)
        assert points.ds.shape == (10000, 2)  # One less column here due to geometry storing X and Y
        assert all(np.isfinite(points_arr[:, 2]))

        # The output should respect the default band naming and the input CRS
        assert np.array_equal(points.ds.columns, ["b1", "geometry"])
        assert points.crs == img1.crs

        # Try setting the band name
        points = img1.to_pointcloud(data_column_name="lol", subsample=10)
        assert np.array_equal(points.ds.columns, ["lol", "geometry"])

        # Keeping the nodata values
        points_invalid = img1.to_pointcloud(subsample=10000, random_state=42, skip_nodata=False)

        # The subsampled values should not all be valid and the right shape
        assert points_invalid.ds.shape == (10000, 2)  # One less column here due to geometry storing X and Y
        assert any(~np.isfinite(points_invalid["b1"].values))

        # 4/ Multi-band real raster
        img2 = gu.Raster(self.landsat_rgb_path)

        # By default only loads a single band without loading
        points_arr = img2.to_pointcloud(subsample=10, as_array=True)
        points = img2.to_pointcloud(subsample=10)

        assert points_arr.shape == (10, 3)
        assert points.ds.shape == (10, 2)  # One less column here due to geometry storing X and Y
        assert not img2.is_loaded

        # Storing auxiliary bands
        points_arr = img2.to_pointcloud(subsample=10, as_array=True, auxiliary_data_bands=[2, 3])
        points = img2.to_pointcloud(subsample=10, auxiliary_data_bands=[2, 3])
        assert points_arr.shape == (10, 5)
        assert points.ds.shape == (10, 4)  # One less column here due to geometry storing X and Y
        assert not img2.is_loaded
        assert np.array_equal(points.ds.columns, ["b1", "b2", "b3", "geometry"])

        # Try setting the column name of a specific band while storing all
        points = img2.to_pointcloud(subsample=10, data_column_name="yes", data_band=2, auxiliary_data_bands=[1, 3])
        assert np.array_equal(points.ds.columns, ["yes", "b1", "b3", "geometry"])

        # 5/ Error raising
        with pytest.raises(ValueError, match="Data column name must be a string.*"):
            img1.to_pointcloud(data_column_name=1)  # type: ignore
        with pytest.raises(
            ValueError,
            match=re.escape("Data band number must be an integer between 1 and the total number of bands (3)."),
        ):
            img2.to_pointcloud(data_band=4)
        with pytest.raises(
            ValueError, match="Passing auxiliary column names requires passing auxiliary data band numbers as well."
        ):
            img2.to_pointcloud(auxiliary_column_names=["a"])
        with pytest.raises(
            ValueError, match="Auxiliary data band number must be an iterable containing only integers."
        ):
            img2.to_pointcloud(auxiliary_data_bands=[1, 2.5])  # type: ignore
            img2.to_pointcloud(auxiliary_data_bands="lol")  # type: ignore
        with pytest.raises(
            ValueError,
            match=re.escape("Auxiliary data band numbers must be between 1 and the total number of bands (3)."),
        ):
            img2.to_pointcloud(auxiliary_data_bands=[0])
            img2.to_pointcloud(auxiliary_data_bands=[4])
        with pytest.raises(
            ValueError, match=re.escape("Main data band 1 should not be listed in auxiliary data bands [1, 2].")
        ):
            img2.to_pointcloud(auxiliary_data_bands=[1, 2])
        with pytest.raises(ValueError, match="Auxiliary column names must be an iterable containing only strings."):
            img2.to_pointcloud(auxiliary_data_bands=[2, 3], auxiliary_column_names=["lol", 1])
        with pytest.raises(
            ValueError, match="Length of auxiliary column name and data band numbers should be the same*"
        ):
            img2.to_pointcloud(auxiliary_data_bands=[2, 3], auxiliary_column_names=["lol", "lol2", "lol3"])

    def test_from_pointcloud(self) -> None:
        """Test from_pointcloud method."""

        # 1/ Create a small raster to test point sampling on
        shape = (5, 5)
        nodata = 100
        img_arr = np.arange(np.prod(shape), dtype="int32").reshape(shape)
        transform = rio.transform.from_origin(0, 5, 1, 1)
        img1 = gu.Raster.from_array(img_arr, transform=transform, crs=4326, nodata=nodata)

        # Check both inputs work (grid coords or transform+shape) on a subsample
        pc1 = img1.to_pointcloud(subsample=10)
        img1_sub = gu.Raster.from_pointcloud_regular(pc1, transform=transform, shape=shape)

        grid_coords1 = img1.coords(grid=False)
        img1_sub2 = gu.Raster.from_pointcloud_regular(pc1, grid_coords=grid_coords1)

        assert img1_sub.raster_equal(img1_sub2)

        # Check that number of valid values are equal to point cloud size
        assert np.count_nonzero(~img1_sub.data.mask) == 10

        # With no subsampling, should get the exact same raster back
        pc1_full = img1.to_pointcloud()
        img1_full = gu.Raster.from_pointcloud_regular(pc1_full, transform=transform, shape=shape, nodata=nodata)
        assert img1.raster_equal(img1_full, warn_failure_reason=True)

        # 2/ Single-band real raster with nodata values
        img2 = gu.Raster(self.aster_dem_path)
        nodata = img2.nodata
        transform = img2.transform
        shape = img2.shape

        # Check both inputs work (grid coords or transform+shape) on a subsample
        pc2 = img2.to_pointcloud(subsample=10000, random_state=42)
        img2_sub = gu.Raster.from_pointcloud_regular(pc2, transform=transform, shape=shape, nodata=nodata)

        grid_coords2 = img2.coords(grid=False)
        img2_sub2 = gu.Raster.from_pointcloud_regular(pc2, grid_coords=grid_coords2, nodata=nodata)

        assert img2_sub.raster_equal(img2_sub2, warn_failure_reason=True)

        # Check that number of valid values are equal to point cloud size
        assert np.count_nonzero(~img2_sub.data.mask) == 10000

        # With no subsampling, should get the exact same raster back
        pc2_full = img2.to_pointcloud()
        img2_full = gu.Raster.from_pointcloud_regular(pc2_full, transform=transform, shape=shape, nodata=nodata)
        assert img2.raster_equal(img2_full, warn_failure_reason=True, strict_masked=False)

        # 3/ Error raising
        with pytest.raises(TypeError, match="Input grid coordinates must be 1D arrays.*"):
            gu.Raster.from_pointcloud_regular(pc1, grid_coords=(1, "lol"))  # type: ignore
        with pytest.raises(ValueError, match="Grid coordinates must be regular*"):
            grid_coords1[0][0] += 1
            gu.Raster.from_pointcloud_regular(pc1, grid_coords=grid_coords1)  # type: ignore
        with pytest.raises(
            ValueError, match="Either grid coordinates or both geotransform and shape must be provided."
        ):
            gu.Raster.from_pointcloud_regular(pc1)



"""
Test projtools
"""

import os.path

import geopandas as gpd
import numpy as np
import pyproj.exceptions
import pytest
from shapely.geometry import Point, Polygon

import geoutils as gu
import geoutils.projtools as pt
from geoutils import examples


class TestProjTools:
    landsat_b4_path = examples.get_path("everest_landsat_b4")
    landsat_b4_crop_path = examples.get_path("everest_landsat_b4_cropped")
    landsat_rgb_path = examples.get_path("everest_landsat_rgb")
    everest_outlines_path = examples.get_path("everest_rgi_outlines")
    aster_dem_path = examples.get_path("exploradores_aster_dem")
    aster_outlines_path = examples.get_path("exploradores_rgi_outlines")

    def test_latlon_to_utm(self) -> None:
        # First: Check errors are raised when format is invalid

        # If format is invalid
        with pytest.raises(TypeError):
            pt.latlon_to_utm("1", 100)  # type: ignore
        # If values are outside limits: latitude above or below 90
        with pytest.raises(ValueError):
            pt.latlon_to_utm(91, 0)
        with pytest.raises(ValueError):
            pt.latlon_to_utm(-91, 0)
        # Or longitude above or below 180
        with pytest.raises(ValueError):
            pt.latlon_to_utm(0, -181)
        with pytest.raises(ValueError):
            pt.latlon_to_utm(0, 181)

        # Second: check that the UTM zone is correct
        # Lower left belongs to the zone, so 0, 0 should be in zone 31N
        assert pt.latlon_to_utm(0, 0) == "30N"
        # Test extreme zones
        assert pt.latlon_to_utm(-79, -179) == "1S"
        assert pt.latlon_to_utm(79, -179) == "1N"
        assert pt.latlon_to_utm(-79, 179) == "60S"
        assert pt.latlon_to_utm(79, 179) == "60N"
        # Test some middles zones
        assert pt.latlon_to_utm(1, -59) == "21N"
        assert pt.latlon_to_utm(1, 61) == "41N"
        assert pt.latlon_to_utm(-1, -121) == "10S"
        assert pt.latlon_to_utm(-1, 119) == "50S"

        # Third, check that any floating or integer type works
        assert pt.latlon_to_utm(0.1, 0.1)
        assert pt.latlon_to_utm(np.float32(0.1), np.float32(0.1))
        assert pt.latlon_to_utm(np.int16(0), np.int16(0))

    def test_utm_to_epsg(self) -> None:
        """Check that the EPSG codes derived from UTM zones are correct"""

        # First: Check errors are raised when format is invalid

        # If there isn't 2 digits for the code
        with pytest.raises(pyproj.exceptions.CRSError):
            pt.utm_to_epsg("100N")
        # If type is incorrect
        with pytest.raises(TypeError):
            pt.utm_to_epsg({"utm": "10N"})  # type: ignore
        # If the code digits does not exist
        with pytest.raises(pyproj.exceptions.CRSError):
            pt.utm_to_epsg("61N")
        # If the north-south zone letter is incorrect
        with pytest.raises(pyproj.exceptions.CRSError):
            pt.utm_to_epsg("61E")

        # Second: Check that the EPSG code is correct
        # https://epsg.io/32601
        assert pt.utm_to_epsg("1N") == 32601
        # https://epsg.io/32701
        assert pt.utm_to_epsg("1S") == 32701
        # https://epsg.io/32&660
        assert pt.utm_to_epsg("60N") == 32660
        # https://epsg.io/32760
        assert pt.utm_to_epsg("60S") == 32760

        # Third: Check that different format work: single digit, lower-case
        assert pt.utm_to_epsg("1N") == pt.utm_to_epsg("01N") == pt.utm_to_epsg("01n")
        assert pt.utm_to_epsg("08s") == pt.utm_to_epsg("8S") == pt.utm_to_epsg("08S")

    @pytest.mark.parametrize("example", [landsat_b4_path, aster_dem_path])  # type: ignore
    def test_get_metric_crs_utm(self, example: str) -> None:
        """Check that the function works consistently with GeoPandas for UTM."""

        # Open raster
        rast = gu.Raster(example)

        # Get utm zone from GeoPandas and GeoUtils
        crs_gu = rast.get_metric_crs()
        crs_geopandas = rast.get_metric_crs(method="geopandas")

        # Verify they are the same
        assert crs_gu == crs_geopandas

    def test_get_metric_crs_ups(self) -> None:
        """Check that the function works for UPS with points at high latitude."""

        # Create a vector of a single point in geographic coordinates
        point_north = Point([0, 84])
        point_south = Point([0, -84])
        vect_north = gu.Vector(gpd.GeoDataFrame({"geometry": [point_north]}, crs=pyproj.CRS.from_epsg(4326)))
        vect_south = gu.Vector(gpd.GeoDataFrame({"geometry": [point_south]}, crs=pyproj.CRS.from_epsg(4326)))

        # Check that UPS North and South are indeed returned
        assert vect_north.get_metric_crs() == pyproj.CRS.from_epsg(32661)
        assert vect_south.get_metric_crs() == pyproj.CRS.from_epsg(32761)

    @pytest.mark.parametrize("example", [landsat_b4_path, aster_dem_path])  # type: ignore
    def test_latlon_reproject(self, example: str) -> None:
        """
        Check that to and from latlon projections are self consistent within tolerated rounding errors
        """

        img = gu.Raster(example)

        # Test on random points
        nsample = 100
        rng = np.random.default_rng(42)
        randx = rng.integers(low=img.bounds.left, high=img.bounds.right, size=(nsample,))
        randy = rng.integers(low=img.bounds.bottom, high=img.bounds.top, size=(nsample,))

        lat, lon = pt.reproject_to_latlon([list(randx), list(randy)], img.crs)
        x, y = pt.reproject_from_latlon([lat, lon], img.crs)

        assert np.all(x == randx)
        assert np.all(y == randy)

    def test_merge_bounds(self) -> None:
        """
        Check that merge_bounds and bounds2poly work as expected for all kinds of bounds objects.
        """
        img1 = gu.Raster(self.landsat_b4_path)
        img2 = gu.Raster(self.landsat_b4_crop_path)

        # Check union (default) - with Raster objects
        out_bounds = pt.merge_bounds((img1, img2))
        assert out_bounds[0] == min(img1.bounds.left, img2.bounds.left)
        assert out_bounds[1] == min(img1.bounds.bottom, img2.bounds.bottom)
        assert out_bounds[2] == max(img1.bounds.right, img2.bounds.right)
        assert out_bounds[3] == max(img1.bounds.top, img2.bounds.top)

        # Check intersection - with Raster objects
        out_bounds = pt.merge_bounds((img1, img2), merging_algorithm="intersection")
        assert out_bounds[0] == max(img1.bounds.left, img2.bounds.left)
        assert out_bounds[1] == max(img1.bounds.bottom, img2.bounds.bottom)
        assert out_bounds[2] == min(img1.bounds.right, img2.bounds.right)
        assert out_bounds[3] == min(img1.bounds.top, img2.bounds.top)

        # Check that the results is the same with rio.BoundingBoxes
        out_bounds2 = pt.merge_bounds((img1.bounds, img2.bounds), merging_algorithm="intersection")
        assert out_bounds2 == out_bounds

        # Check that the results is the same with a list
        out_bounds2 = pt.merge_bounds((list(img1.bounds), list(img2.bounds)), merging_algorithm="intersection")
        assert out_bounds2 == out_bounds

        # Check with gpd.GeoDataFrame
        outlines = gu.Vector(gu.examples.get_path("everest_rgi_outlines"))
        outlines = gu.Vector(outlines.ds.to_crs(img1.crs))  # reproject to img1's CRS
        out_bounds = pt.merge_bounds((img1, outlines.ds))

        assert out_bounds[0] == min(img1.bounds.left, outlines.ds.total_bounds[0])
        assert out_bounds[1] == min(img1.bounds.bottom, outlines.ds.total_bounds[1])
        assert out_bounds[2] == max(img1.bounds.right, outlines.ds.total_bounds[2])
        assert out_bounds[3] == max(img1.bounds.top, outlines.ds.total_bounds[3])

    # Try all vectors and rasters
    @pytest.mark.parametrize(
        "fn_raster_or_vector",
        [
            landsat_b4_path,
            landsat_rgb_path,
            landsat_b4_crop_path,
            everest_outlines_path,
            aster_dem_path,
            aster_outlines_path,
        ],
    )  # type: ignore
    # Try with geographic, a UTM zone and a Robinson
    @pytest.mark.parametrize("out_crs", [pyproj.CRS.from_epsg(4326), pyproj.CRS.from_epsg(32610)])  # type: ignore
    @pytest.mark.parametrize("densify_points", [2, 10, 5000])  # type: ignore
    def test_get_footprint_projected(self, fn_raster_or_vector: str, out_crs: pyproj.CRS, densify_points: int) -> None:
        """Test the get footprint projected function."""

        # Open raster or vector
        if os.path.splitext(os.path.basename(fn_raster_or_vector))[-1] == ".tif":
            rast_or_vect = gu.Raster(fn_raster_or_vector)
        else:
            rast_or_vect = gu.Vector(fn_raster_or_vector)

        # Get footprint
        footprint = rast_or_vect.get_footprint_projected(out_crs=out_crs, densify_points=densify_points)

        # Assert it is a vector containing a polygon geometry
        assert isinstance(footprint, gu.Vector)
        assert isinstance(footprint.geometry[0], Polygon)

        # Check that the original corner points were conserved
        left, bottom, right, top = rast_or_vect.bounds
        corners = [Point([x, y]) for (x, y) in [(left, bottom), (left, top), (right, top), (right, bottom)]]
        df = gpd.GeoDataFrame({"geometry": corners}, crs=rast_or_vect.crs)
        df_reproj = df.to_crs(crs=out_crs)

        assert all(point.coords[0] in footprint.geometry[0].exterior.coords[:] for point in df_reproj.geometry)

        # Check that densification yields a logical amount of points
        # (4 initial corner points times the densification factor + the last point)
        assert len(footprint.geometry[0].exterior.coords[:]) == densify_points * 4 + 1



"""
Python arithmetic
=================

This example demonstrates arithmetic operations using raster arithmetic on :class:`Rasters<geoutils.Raster>`. See :ref:`core-py-ops` for more details.
"""

# %%
# We open a raster

# sphinx_gallery_thumbnail_number = 2
import geoutils as gu

filename_rast = gu.examples.get_path("everest_landsat_b4")
rast = gu.Raster(filename_rast)
rast

# %% We plot the original raster.
rast.plot(cmap="Greys_r")

# %%
# Performing arithmetic operations implicitly loads the data.
rast = (rast + 1.0) ** 0.5 / 5
rast.plot(cmap="Greys_r")

# %%
#
# .. important::
#        Arithmetic operations cast to new :class:`dtypes<numpy.dtype>` automatically following NumPy coercion rules. If we had written ``(rast + 1)``,
#        this calculation would have conserved the original :class:`numpy.uint8` :class:`dtype<numpy.dtype>` of the raster.
#
# Logical comparison operations will naturally cast to a :class:`Mask<geoutils.Mask>`.

mask = rast == 200
mask

# %%
# :class:`Masks<geoutils.Mask>` support python logical operators to be combined together

mask = (rast >= 3) | (rast % 2 == 0) & (rast != 80)
mask.plot()

# %%
# Finally, :class:`Masks<geoutils.Mask>` can be used for indexing and assigning to a :class:`Rasters<geoutils.Raster>`

values = rast[mask]



# Copyright (c) 2025 GeoUtils developers
#
# This file is part of the GeoUtils project:
# https://github.com/glaciohack/geoutils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Typing aliases for internal use."""

from __future__ import annotations

import sys
from typing import Any, List, Tuple, Union

import numpy as np

# Only for Python >= 3.9
if sys.version_info.minor >= 9:

    from numpy.typing import (  # this syntax works starting on Python 3.9
        ArrayLike,
        DTypeLike,
        NDArray,
    )

    # Mypy has issues with the builtin Number type (https://github.com/python/mypy/issues/3186)
    Number = Union[int, float, np.integer[Any], np.floating[Any]]

    # Simply define here if they exist
    DTypeLike = DTypeLike
    ArrayLike = ArrayLike

    # Use NDArray wrapper to easily define numerical (float or int) N-D array types, and boolean N-D array types
    NDArrayNum = NDArray[Union[np.floating[Any], np.integer[Any]]]
    NDArrayBool = NDArray[np.bool_]
    # Define numerical (float or int) masked N-D array type
    MArrayNum = np.ma.masked_array[Any, np.dtype[Union[np.floating[Any], np.integer[Any]]]]
    MArrayBool = np.ma.masked_array[Any, np.dtype[np.bool_]]

# For backward compatibility before Python 3.9
else:

    # Mypy has issues with the builtin Number type (https://github.com/python/mypy/issues/3186)
    Number = Union[int, float, np.integer, np.floating]  # type: ignore

    # Make an array-like type (since the array-like numpy type only exists in numpy>=1.20)
    DTypeLike = Union[str, type, np.dtype]  # type: ignore
    ArrayLike = Union[np.ndarray, np.ma.masked_array, List[Any], Tuple[Any]]  # type: ignore

    # Define generic types for NumPy array and masked-array (behaves as "Any" before 3.9 and plugin)
    NDArrayNum = np.ndarray  # type: ignore
    NDArrayBool = np.ndarray  # type: ignore
    MArrayNum = np.ma.masked_array  # type: ignore
    MArrayBool = np.ma.masked_array  # type: ignore



# Copyright (c) 2025 GeoUtils developers
#
# This file is part of the GeoUtils project:
# https://github.com/glaciohack/geoutils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Array tools related to rasters."""

from __future__ import annotations

import warnings

import numpy as np

import geoutils as gu
from geoutils._typing import MArrayNum, NDArrayBool, NDArrayNum


def get_mask_from_array(array: NDArrayNum | NDArrayBool | MArrayNum) -> NDArrayBool:
    """
    Return the mask of invalid values, whether array is a ndarray with NaNs or a np.ma.masked_array.

    :param array: Input array.

    :returns invalid_mask: boolean array, True where array is masked or Nan.
    """
    mask = (array.mask | ~np.isfinite(array.data)) if isinstance(array, np.ma.masked_array) else ~np.isfinite(array)
    return mask.squeeze()


def get_array_and_mask(
    array: NDArrayNum | MArrayNum, check_shape: bool = True, copy: bool = True
) -> tuple[NDArrayNum, NDArrayBool]:
    """
    Return array with masked values set to NaN and the associated mask.
    Works whether array is a ndarray with NaNs or a np.ma.masked_array.

    :param array: Input array.
    :param check_shape: Validate that the array is either a 1D array, a 2D array or a 3D array of shape (1, rows, cols).
    :param copy: Return a copy of 'array'. If False, a view will be attempted (and warn if not possible)

    :returns array_data, invalid_mask: a tuple of ndarrays. First is array with invalid pixels converted to NaN, \
    second is mask of invalid pixels (True if invalid).
    """
    #
    if isinstance(array, gu.Raster):
        array = array.data

    if check_shape:
        if array.ndim > 2 and array.shape[0] > 1:
            raise ValueError(
                f"Invalid array shape given: {array.shape}." "Expected 2D array or 3D array where arr.shape[0] == 1"
            )

    # If an occupied mask exists and a view was requested, trigger a warning.
    if not copy and np.any(getattr(array, "mask", False)):
        warnings.warn("Copying is required to respect the mask. Returning copy. Set 'copy=True' to hide this message.")
        copy = True

    # If array is of type integer and has a mask, it needs to be converted to float (to assign nans)
    if np.any(getattr(array, "mask", False)) and np.issubdtype(array.dtype, np.integer):  # type: ignore
        array = array.astype(np.float32)  # type: ignore

    # Convert into a regular ndarray (a view or copy depending on the 'copy' argument)
    array_data = np.array(array).squeeze() if copy else np.asarray(array).squeeze()

    # Get the mask of invalid pixels and set nans if it is occupied.
    invalid_mask = get_mask_from_array(array)
    if np.any(invalid_mask):
        array_data[invalid_mask] = np.nan

    return array_data, invalid_mask


def get_valid_extent(array: NDArrayNum | NDArrayBool | MArrayNum) -> tuple[int, ...]:
    """
    Return (rowmin, rowmax, colmin, colmax), the first/last row/column of array with valid pixels
    """
    if not array.dtype == "bool":
        valid_mask = ~get_mask_from_array(array)
    else:
        # Not sure why Mypy is not recognizing that the type of the array can only be bool here
        valid_mask = array  # type: ignore
    cols_nonzero = np.where(np.count_nonzero(valid_mask, axis=0) > 0)[0]
    rows_nonzero = np.where(np.count_nonzero(valid_mask, axis=1) > 0)[0]
    return rows_nonzero[0], rows_nonzero[-1], cols_nonzero[0], cols_nonzero[-1]


def get_xy_rotated(raster: gu.Raster, along_track_angle: float) -> tuple[NDArrayNum, NDArrayNum]:
    """
    Rotate x, y axes of image to get along- and cross-track distances.
    :param raster: Raster to get x,y positions from.
    :param along_track_angle: Angle by which to rotate axes (degrees)

    :returns xxr, yyr: Arrays corresponding to along (x) and cross (y) track distances.
    """

    myang = np.deg2rad(along_track_angle)

    # Get grid coordinates
    # (only relative is important, we don't care about offsets, so let's fix lower-left to make the tests easier
    # by starting nicely at 0,0)
    xx, yy = raster.coords(grid=True, force_offset="ll")
    xx -= np.min(xx)
    yy -= np.min(yy)

    # Get rotated coordinates

    # For along-track
    xxr = xx * np.cos(myang) - yy * np.sin(myang)
    # For cross-track
    yyr = xx * np.sin(myang) + yy * np.cos(myang)

    # Re-initialize coordinate at zero
    xxr -= np.nanmin(xxr)
    yyr -= np.nanmin(yyr)

    return xxr, yyr



# Copyright (c) 2025 GeoUtils developers
#
# This file is part of the GeoUtils project:
# https://github.com/glaciohack/geoutils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Functions for manipulating georeferencing of the raster objects.
"""

from __future__ import annotations

import warnings
from typing import Iterable, Literal

import numpy as np
import rasterio as rio

from geoutils._config import config
from geoutils._typing import ArrayLike, DTypeLike, NDArrayNum


def _ij2xy(
    i: ArrayLike,
    j: ArrayLike,
    transform: rio.transform.Affine,
    area_or_point: Literal["Area", "Point"] | None,
    shift_area_or_point: bool | None = None,
    force_offset: str | None = None,
) -> tuple[NDArrayNum, NDArrayNum]:
    """See description of Raster.ij2xy."""

    # If undefined, default to the global system config
    if shift_area_or_point is None:
        shift_area_or_point = config["shift_area_or_point"]

    # Shift by half a pixel back for "Point" interpretation
    if shift_area_or_point and force_offset is None:
        if area_or_point is not None and area_or_point == "Point":
            i = np.asarray(i) - 0.5
            j = np.asarray(j) - 0.5

    # Default offset is upper-left for raster coordinates
    if force_offset is None:
        force_offset = "ul"

    x, y = rio.transform.xy(transform, i, j, offset=force_offset)

    return x, y


def _xy2ij(
    x: ArrayLike,
    y: ArrayLike,
    transform: rio.transform.Affine,
    area_or_point: Literal["Area", "Point"] | None,
    op: type = np.float32,
    precision: float | None = None,
    shift_area_or_point: bool | None = None,
) -> tuple[NDArrayNum, NDArrayNum]:
    """See description of Raster.xy2ij."""

    # If undefined, default to the global system config
    if shift_area_or_point is None:
        shift_area_or_point = config["shift_area_or_point"]

    # Input checks
    if op not in [np.float32, np.float64, float]:
        raise UserWarning(
            "Operator is not of type float: rio.Dataset.index might "
            "return unreliable indexes due to rounding issues."
        )

    i, j = rio.transform.rowcol(transform, x, y, op=op, precision=precision)

    # Necessary because rio.Dataset.index does not return abc.Iterable for a single point
    if not isinstance(i, Iterable):
        i, j = (
            np.asarray(
                [
                    i,
                ]
            ),
            np.asarray(
                [
                    j,
                ]
            ),
        )
    else:
        i, j = (np.asarray(i), np.asarray(j))

    # AREA_OR_POINT GDAL attribute, i.e. does the value refer to the upper left corner "Area" or
    # the center of pixel "Point". This normally has no influence on georeferencing, it's only
    # about the interpretation of the raster values, and thus can affect sub-pixel interpolation,
    # for more details see: https://gdal.org/user/raster_data_model.html#metadata

    # If the user wants to shift according to the interpretation
    if shift_area_or_point:

        # Shift by half a pixel if the AREA_OR_POINT attribute is "Point", otherwise leave as is
        if area_or_point is not None and area_or_point == "Point":
            if not isinstance(i.flat[0], (np.floating, float)):
                raise ValueError("Operator must return np.floating values to perform pixel interpretation shifting.")

            i += 0.5
            j += 0.5

    # Convert output indexes to integer if they are all whole numbers
    if np.all(np.mod(i, 1) == 0) and np.all(np.mod(j, 1) == 0):
        i = i.astype(int)
        j = j.astype(int)

    return i, j


def _coords(
    transform: rio.transform.Affine,
    shape: tuple[int, int],
    area_or_point: Literal["Area", "Point"] | None,
    grid: bool = True,
    shift_area_or_point: bool | None = None,
    force_offset: str | None = None,
) -> tuple[NDArrayNum, NDArrayNum]:
    """See description of Raster.coords."""

    # The coordinates are extracted from indexes 0 to shape
    _, yy = _ij2xy(
        i=np.arange(shape[0] - 1, -1, -1),
        j=0,
        transform=transform,
        area_or_point=area_or_point,
        shift_area_or_point=shift_area_or_point,
        force_offset=force_offset,
    )
    xx, _ = _ij2xy(
        i=0,
        j=np.arange(shape[1]),
        transform=transform,
        area_or_point=area_or_point,
        shift_area_or_point=shift_area_or_point,
        force_offset=force_offset,
    )

    # If grid is True, return coordinate grids
    if grid:
        meshgrid = tuple(np.meshgrid(xx, np.flip(yy)))
        return meshgrid  # type: ignore
    else:
        return np.asarray(xx), np.asarray(yy)


def _outside_image(
    xi: ArrayLike,
    yj: ArrayLike,
    transform: rio.transform.Affine,
    shape: tuple[int, int],
    area_or_point: Literal["Area", "Point"] | None,
    index: bool = True,
) -> bool:
    """See description of Raster.outside_image."""

    if not index:
        xi, xj = _xy2ij(xi, yj, transform=transform, area_or_point=area_or_point)

    if np.any(np.array((xi, yj)) < 0):
        return True
    elif np.asanyarray(xi) > shape[1] or np.asanyarray(yj) > shape[0]:
        return True
    else:
        return False


def _res(transform: rio.transform.Affine) -> tuple[float, float]:
    """See description of Raster.res"""

    return transform[0], abs(transform[4])


def _bounds(transform: rio.transform.Affine, shape: tuple[int, int]) -> rio.coords.BoundingBox:
    """See description of Raster.bounds."""

    return rio.coords.BoundingBox(*rio.transform.array_bounds(height=shape[0], width=shape[1], transform=transform))


def _cast_pixel_interpretation(
    area_or_point1: Literal["Area", "Point"] | None, area_or_point2: Literal["Area", "Point"] | None
) -> Literal["Area", "Point"] | None:
    """
    Cast two pixel interpretations and warn if not castable.

    Casts to:
     - "Area" if both are "Area",
     - "Point" if both are "Point",
     - None if any of the interpretation is None, or
     - None if one is "Area" and the other "Point" (and raises a warning).
    """

    # If one is None, cast to None
    if area_or_point1 is None or area_or_point2 is None:
        area_or_point_out = None
    # If both are equal and not None
    elif area_or_point1 == area_or_point2:
        area_or_point_out = area_or_point1
    else:
        area_or_point_out = None
        msg = (
            'One raster has a pixel interpretation "Area" and the other "Point". To silence this warning, '
            "either correct the pixel interpretation of one raster, or deactivate "
            'warnings of pixel interpretation with geoutils.config["warn_area_or_point"]=False.'
        )
        if config["warn_area_or_point"]:
            warnings.warn(message=msg, category=UserWarning)

    return area_or_point_out


# Function to set the default nodata values for any given dtype
# Similar to GDAL for int types, but without absurdly long nodata values for floats.
# For unsigned types, the maximum value is chosen (with a max of 99999).
# For signed types, the minimum value is chosen (with a min of -99999).
def _default_nodata(dtype: DTypeLike) -> int:
    """
    Set the default nodata value for any given dtype, when this is not provided.
    """
    default_nodata_lookup = {
        "uint8": 255,
        "int8": -128,
        "uint16": 65535,
        "int16": -32768,
        "uint32": 99999,
        "int32": -99999,
        "float16": -99999,
        "float32": -99999,
        "float64": -99999,
        "float128": -99999,
        "longdouble": -99999,  # This is float64 on Windows, float128 on other systems, for compatibility
    }
    # Check argument dtype is as expected
    if not isinstance(dtype, (str, np.dtype, type)):
        raise TypeError(f"dtype {dtype} not understood.")

    # Convert numpy types to string
    if isinstance(dtype, type):
        dtype = np.dtype(dtype).name

    # Convert np.dtype to string
    if isinstance(dtype, np.dtype):
        dtype = dtype.name

    if dtype in default_nodata_lookup.keys():
        return default_nodata_lookup[dtype]
    else:
        raise NotImplementedError(f"No default nodata value set for dtype {dtype}.")


def _cast_nodata(out_dtype: DTypeLike, nodata: int | float | None) -> int | float | None:
    """
    Cast nodata value for output data type to default nodata if incompatible.

    :param out_dtype: Dtype of output array.
    :param nodata: Nodata value.

    :return: Cast nodata value.
    """

    if out_dtype == bool:
        nodata = None
    if nodata is not None and not rio.dtypes.can_cast_dtype(nodata, out_dtype):
        nodata = _default_nodata(out_dtype)
    else:
        nodata = nodata

    return nodata



"""
From/to GeoPandas
=================

This example demonstrates importing or exporting a :class:`geopandas.GeoDataFrame` from and to a :class:`~geoutils.Vector`.
"""

# %%
# A vector can be imported from a :class:`geopandas.GeoDataFrame` simply by instantiating :class:`~geoutils.Vector`.

import geopandas as gpd

import geoutils as gu

filename_vect = gu.examples.get_path("exploradores_rgi_outlines")
ds = gpd.read_file(filename_vect)
vect = gu.Vector(ds)
vect

# %%
# We plot the vector.

vect.plot(column="RGIId", add_cbar=False)

# %%
# To export, the :class:`geopandas.GeoDataFrame` is always stored as an attribute as :class:`~geoutils.Vector` is composed from it. See :ref:`core-composition`.

vect.ds



"""
Reduction from window
=====================

This example demonstrates the reduction of windowed raster values around a point using :func:`~geoutils.Raster.value_at_coords`.
"""

# %%
# We open an example raster, a digital elevation model in South America.

# sphinx_gallery_thumbnail_number = 3
import geoutils as gu

filename_rast = gu.examples.get_path("exploradores_aster_dem")
rast = gu.Raster(filename_rast)
rast = rast.crop([rast.bounds.left, rast.bounds.bottom, rast.bounds.left + 2000, rast.bounds.bottom + 2000])

# Plot the raster
rast.plot(cmap="terrain")

# %%
# We generate a random subsample of 100 coordinates to extract.

import geopandas as gpd
import numpy as np

# Replace by Raster function once done
rng = np.random.default_rng(42)
x_coords = rng.uniform(rast.bounds.left + 50, rast.bounds.right - 50, 50)
y_coords = rng.uniform(rast.bounds.bottom + 50, rast.bounds.top - 50, 50)

vals = rast.reduce_points((x_coords, y_coords))

# %%
# Replace by Vector function once done
ds = gpd.GeoDataFrame(geometry=gpd.points_from_xy(x=x_coords, y=y_coords), crs=rast.crs)
ds["vals"] = vals
ds.plot(column="vals", cmap="terrain", legend=True, vmin=np.nanmin(rast), vmax=np.nanmax(rast))

# %%
# By default, :func:`~geoutils.Raster.value_at_coords` extracts the closest pixel value. But it can also be passed a window size and reductor function to
# extract an average value or other statistic based on neighbouring pixels.

vals_reduced = rast.reduce_points((x_coords, y_coords), window=5, reducer_function=np.nanmedian)

np.nanmean(vals - vals_reduced)

# %%
# The mean difference in extracted values is quite significant at 0.3 meters!
# We can visualize how the sampling took place in window.

# Replace by Vector fonction once done
coords = rast.coords(grid=True)
x_closest = rast.copy(new_array=coords[0]).reduce_points((x_coords, y_coords)).squeeze()
y_closest = rast.copy(new_array=coords[1]).reduce_points((x_coords, y_coords)).squeeze()
from shapely.geometry import box

geometry = [
    box(x - 2 * rast.res[0], y - 2 * rast.res[1], x + 2 * rast.res[0], y + 2 * rast.res[1])
    for x, y in zip(x_closest, y_closest)
]
ds = gpd.GeoDataFrame(geometry=geometry, crs=rast.crs)
ds["vals"] = vals_reduced
ds.plot(column="vals", cmap="terrain", legend=True, vmin=np.nanmin(rast), vmax=np.nanmax(rast))



"""Tests for geometry operations on vectors."""

from __future__ import annotations

import geopandas as gpd
import matplotlib.pyplot as plt
import numpy as np
import pytest
from geopandas.testing import assert_geodataframe_equal
from shapely import LineString, MultiLineString, MultiPolygon, Polygon

import geoutils as gu
from geoutils.vector.geometric import _extract_vertices, _generate_voronoi_polygons


class TestGeometric:

    # Create a synthetic vector file with a square of size 1, started at position (10, 10)
    poly1 = Polygon([(10, 10), (11, 10), (11, 11), (10, 11)])
    gdf = gpd.GeoDataFrame({"geometry": [poly1]}, crs="EPSG:4326")
    vector = gu.Vector(gdf)

    # Same with a square started at position (5, 5)
    poly2 = Polygon([(5, 5), (6, 5), (6, 6), (5, 6)])
    gdf = gpd.GeoDataFrame({"geometry": [poly2]}, crs="EPSG:4326")
    vector2 = gu.Vector(gdf)

    # Create a multipolygon with both
    multipoly = MultiPolygon([poly1, poly2])
    gdf = gpd.GeoDataFrame({"geometry": [multipoly]}, crs="EPSG:4326")
    vector_multipoly = gu.Vector(gdf)

    # Create a synthetic vector file with a square of size 5, started at position (8, 8)
    poly3 = Polygon([(8, 8), (13, 8), (13, 13), (8, 13)])
    gdf = gpd.GeoDataFrame({"geometry": [poly3]}, crs="EPSG:4326")
    vector_5 = gu.Vector(gdf)

    # Create a synthetic LineString geometry
    lines = LineString([(10, 10), (11, 10), (11, 11)])
    gdf = gpd.GeoDataFrame({"geometry": [lines]}, crs="EPSG:4326")
    vector_lines = gu.Vector(gdf)

    # Create a synthetic MultiLineString geometry
    multilines = MultiLineString([[(10, 10), (11, 10), (11, 11)], [(5, 5), (6, 5), (6, 6)]])
    gdf = gpd.GeoDataFrame({"geometry": [multilines]}, crs="EPSG:4326")
    vector_multilines = gu.Vector(gdf)

    def test_extract_vertices(self) -> None:
        """
        Test that extract_vertices works with simple geometries.
        """
        # Polygons
        vertices = _extract_vertices(self.vector.ds)
        assert len(vertices) == 1
        assert vertices == [[(10.0, 10.0), (11.0, 10.0), (11.0, 11.0), (10.0, 11.0), (10.0, 10.0)]]

        # MultiPolygons
        vertices = _extract_vertices(self.vector_multipoly.ds)
        assert len(vertices) == 2
        assert vertices[0] == [(10.0, 10.0), (11.0, 10.0), (11.0, 11.0), (10.0, 11.0), (10.0, 10.0)]
        assert vertices[1] == [(5.0, 5.0), (6.0, 5.0), (6.0, 6.0), (5.0, 6.0), (5.0, 5.0)]

        # LineString
        vertices = _extract_vertices(self.vector_lines.ds)
        assert len(vertices) == 1
        assert vertices == [[(10.0, 10.0), (11.0, 10.0), (11.0, 11.0)]]

        # MultiLineString
        vertices = _extract_vertices(self.vector_multilines.ds)
        assert len(vertices) == 2
        assert vertices[0] == [(10.0, 10.0), (11.0, 10.0), (11.0, 11.0)]
        assert vertices[1] == [(5.0, 5.0), (6.0, 5.0), (6.0, 6.0)]

    def test_generate_voronoi(self) -> None:
        """
        Check that vector.generate_voronoi_polygons works on a simple Polygon.
        Does not work with simple shapes as squares or triangles as the diagram is infinite.
        For now, test on a set of two squares.
        """
        # Check with a multipolygon
        voronoi = _generate_voronoi_polygons(self.vector_multipoly.ds)
        assert len(voronoi) == 2
        vertices = _extract_vertices(voronoi)
        assert vertices == [
            [(5.5, 10.5), (10.5, 10.5), (10.5, 5.5), (5.5, 10.5)],
            [(5.5, 10.5), (10.5, 5.5), (5.5, 5.5), (5.5, 10.5)],
        ]

        # Check that it fails with proper error for too simple geometries
        expected_message = "Invalid geometry, cannot generate finite Voronoi polygons"
        with pytest.raises(ValueError, match=expected_message):
            voronoi = _generate_voronoi_polygons(self.vector.ds)

    def test_buffer_metric(self) -> None:
        """Check that metric buffering works"""

        # Case with two squares: test that the buffered area is without deformations
        # https://epsg.io/32631
        utm31_x_center = 500000
        utm31_y_center = 4649776
        poly1_utm31 = Polygon(
            [
                (utm31_x_center, utm31_y_center),
                (utm31_x_center + 1, utm31_y_center),
                (utm31_x_center + 1, utm31_y_center + 1),
                (utm31_x_center, utm31_y_center + 1),
            ]
        )

        poly2_utm31 = Polygon(
            [
                (utm31_x_center + 10, utm31_y_center + 10),
                (utm31_x_center + 11, utm31_y_center + 10),
                (utm31_x_center + 11, utm31_y_center + 11),
                (utm31_x_center + 10, utm31_y_center + 11),
            ]
        )

        # We initiate the squares of size 1x1 in a UTM projection
        two_squares = gu.Vector(gpd.GeoDataFrame(geometry=[poly1_utm31, poly2_utm31], crs="EPSG:32631"))

        # Their area should now be 1 for each polygon
        assert two_squares.ds.area.values[0] == 1
        assert two_squares.ds.area.values[1] == 1

        # We buffer them
        two_squares_utm_buffered = two_squares.buffer_metric(buffer_size=1.0)

        # Their area should now be 1 (square) + 4 (buffer along the sides) + 4*(pi*1**2 /4)
        # (buffer of corners = quarter-disks)
        expected_area = 1 + 4 + np.pi
        assert two_squares_utm_buffered.ds.area.values[0] == pytest.approx(expected_area, abs=0.01)
        assert two_squares_utm_buffered.ds.area.values[1] == pytest.approx(expected_area, abs=0.01)

        # And the new GeoDataFrame should exactly match that of one buffer from the original one
        direct_gpd_buffer = gu.Vector(
            gpd.GeoDataFrame(geometry=two_squares.ds.buffer(distance=1.0).geometry, crs=two_squares.crs)
        )
        assert_geodataframe_equal(direct_gpd_buffer.ds, two_squares_utm_buffered.ds)

        # Now, if we reproject the original vector in a non-metric system
        two_squares_geographic = gu.Vector(two_squares.ds.to_crs(epsg=4326))
        # We buffer directly the Vector object in the non-metric system
        two_squares_geographic_buffered = two_squares_geographic.buffer_metric(buffer_size=1.0)
        # Then, we reproject that vector in the UTM zone
        two_squares_geographic_buffered_reproj = gu.Vector(
            two_squares_geographic_buffered.ds.to_crs(crs=two_squares.crs)
        )

        # Their area should now be the same as before for each polygon
        assert two_squares_geographic_buffered_reproj.ds.area.values[0] == pytest.approx(expected_area, abs=0.01)
        assert two_squares_geographic_buffered_reproj.ds.area.values[0] == pytest.approx(expected_area, abs=0.01)

        # And this time, it is the reprojected GeoDataFrame that should almost match (within a tolerance of 10e-06)
        assert all(direct_gpd_buffer.ds.geom_equals_exact(two_squares_geographic_buffered_reproj.ds, tolerance=10e-6))

    def test_buffer_without_overlap(self, monkeypatch) -> None:  # type: ignore
        """
        Check that non-overlapping buffer feature works. Does not work on simple geometries, so test on MultiPolygon.
        Yet, very simple geometries yield unexpected results, as is the case for the second test case here.
        """
        # Case 1, test with two squares, in separate Polygons
        two_squares = gu.Vector(gpd.GeoDataFrame(geometry=[self.poly1, self.poly2], crs="EPSG:4326"))

        # Check with buffers that should not overlap
        # ------------------------------------------
        buffer_size = 2
        # We force metric = False, so buffer should raise a GeoPandas warning
        with pytest.warns(UserWarning, match="Geometry is in a geographic CRS.*"):
            buffer = two_squares.buffer_without_overlap(buffer_size, metric=False)

        # Output should be of same size as input and same geometry type
        assert len(buffer.ds) == len(two_squares.ds)
        assert np.all(buffer.ds.geometry.geom_type == two_squares.ds.geometry.geom_type)
        assert buffer.crs == two_squares.crs

        # Extract individual geometries
        polys = []
        for geom in buffer.ds.geometry:
            if geom.geom_type in ["MultiPolygon"]:
                polys.extend(list(geom))
            else:
                polys.append(geom)

        # Check they do not overlap
        for i in range(len(polys)):
            for j in range(i + 1, len(polys)):
                assert not polys[i].intersects(polys[j])

        # buffer should yield the same result as create_mask with buffer, minus the original mask
        mask_nonoverlap = buffer.create_mask(xres=0.1, bounds=(0, 0, 21, 21))
        mask_buffer = two_squares.create_mask(xres=0.1, bounds=(0, 0, 21, 21), buffer=buffer_size)
        mask_nobuffer = two_squares.create_mask(xres=0.1, bounds=(0, 0, 21, 21))
        assert np.all(mask_nobuffer | mask_nonoverlap == mask_buffer)

        # Case 2 - Check with buffers that overlap -> this case is actually not the expected result !
        # -------------------------------
        buffer_size = 5
        # We force metric = False, so buffer should raise a GeoPandas warning
        with pytest.warns(UserWarning, match="Geometry is in a geographic CRS.*"):
            buffer = two_squares.buffer_without_overlap(buffer_size, metric=False)

        # Output should be of same size as input and same geometry type
        assert len(buffer.ds) == len(two_squares.ds)
        assert np.all(buffer.ds.geometry.geom_type == two_squares.ds.geometry.geom_type)

        # Extract individual geometries
        polys = []
        for geom in buffer.ds.geometry:
            if geom.geom_type in ["MultiPolygon"]:
                polys.extend(list(geom))
            else:
                polys.append(geom)

        # Check they do not overlap
        for i in range(len(polys)):
            for j in range(i + 1, len(polys)):
                assert polys[i].intersection(polys[j]).area == 0

        # buffer should yield the same result as create_mask with buffer, minus the original mask
        mask_nonoverlap = buffer.create_mask(xres=0.1, bounds=(0, 0, 21, 21))
        mask_buffer = two_squares.create_mask(xres=0.1, bounds=(0, 0, 21, 21), buffer=buffer_size)
        mask_nobuffer = two_squares.create_mask(xres=0.1, bounds=(0, 0, 21, 21))
        assert np.all(mask_nobuffer | mask_nonoverlap == mask_buffer)

        # Check that plotting runs without errors and close it
        monkeypatch.setattr(plt, "show", lambda: None)
        two_squares.buffer_without_overlap(buffer_size, plot=True)
